import os
import sys

from pathlib import Path

import pandas as pd

import mlinvitrotox.utils.extract_hrms_massbank_data as extract_mbd
import mlinvitrotox.utils.get_true_fps as get_truefps
import mlinvitrotox.utils.get_predicted_fps as get_predfps
import mlinvitrotox.utils.validate_massbank as validate
import mlinvitrotox.utils.training as ml

from mlinvitrotox.constants import (
    # IDs
    CHEM_ID,
    MASSBANK_ID,
    # Paths
    MASSBANK_INPUT_DIR_PATH,
    SIRIUS_INPUT_DIR_PATH,
    INVITRO_INPUT_DIR_PATH,
    MASSBANK_OUTPUT_DIR_PATH,
    FPS_OUTPUT_DIR_PATH,
    INTERMEDIATE_FPS_OUTPUT_DIR_PATH,
    VALIDATION_FPS_OUTPUT_DIR_PATH,
    # Folders
    MASSBANK_INPUT_DIR_FOLDER,  # 1
    SIRIUS_MASSBANK_VALIDATION_DIR_FOLDER,  # 5
    SIRIUS_APPLICATION_DIR_FOLDER,  # 6
    # Files
    MASSBANK_UNIQUE_GUID,  # 0
    SIRIUS_TRAINING_STRUCTURES_INCHI_DTXSID,  # 2
    EXTRACTED_MASSBANK_GUID_DTXSID,  # 10
    EXTRACTED_MASSBANK_SUMMARY_SMILES,  # 11
    EXTRACTED_MASSBANK_SUMMARY_COMPLETE,  # 12
    SIRIUS_FPS_DEFINITIONS,  # 3
    INVITRODB_CHEMICAL_LIST,  # 4
    TRUE_FPS_TRAINING,  # 13
    TRUE_FPS_VALIDATION,  # 14
    PRED_FPS_VALIDATION,  # 15
    PRED_FPS_APPLICATION,  # 16
    VALIDATION_SELECTED_ABSINDEX,  # 21
    VALIDATION_PRED_FPS_DTXSID,  # 20
)

# set input and output directories
massbank_input_directory = Path(MASSBANK_INPUT_DIR_PATH)
sirius_input_directory = Path(SIRIUS_INPUT_DIR_PATH)
invitro_input_directory = Path(INVITRO_INPUT_DIR_PATH)
massbank_output_directory = Path(MASSBANK_OUTPUT_DIR_PATH)
fps_output_directory = Path(FPS_OUTPUT_DIR_PATH)
intermediate_fps_output_directory = Path(INTERMEDIATE_FPS_OUTPUT_DIR_PATH)
validation_fps_output_directory = Path(VALIDATION_FPS_OUTPUT_DIR_PATH)


# * Extract and process MassBank HRMS data

# TODO remove
if 0:
    # process MassBank input files
    massbank_input_folder = massbank_input_directory / MASSBANK_INPUT_DIR_FOLDER
    result_df = extract_mbd.process_files(massbank_input_folder)

    # load SIRIUS training file with INCHIs and DTXSIDs
    sirius_training_file = (
        sirius_input_directory / SIRIUS_TRAINING_STRUCTURES_INCHI_DTXSID
    )
    df_sirius_training = pd.read_csv(sirius_training_file)
    df_sirius_training = df_sirius_training.loc[
        df_sirius_training[CHEM_ID].notna(), [CHEM_ID]
    ]

    # prepare GUIDs
    # (this code only needs to be run once)
    massbank_unique_guid_path = massbank_input_directory / MASSBANK_UNIQUE_GUID
    create_unique_guids = False
    if create_unique_guids:
        n_files = 20000
        extract_mbd.create_and_store_unique_guids(massbank_unique_guid_path, n_files)

    # load unique guids
    df_guid = pd.read_csv(massbank_unique_guid_path)

    # set file paths for output files
    # massbank summary grouped by SMILES
    output_file_structures_validation = (
        massbank_output_directory / EXTRACTED_MASSBANK_SUMMARY_SMILES
    )
    # complete massbank summary by ACCESSION
    output_file_extracted_massbank = (
        massbank_output_directory / EXTRACTED_MASSBANK_SUMMARY_COMPLETE
    )
    # file to match GUID and DTXSID
    output_file_guid_dtxsid = massbank_output_directory / EXTRACTED_MASSBANK_GUID_DTXSID

    # filter data
    extract_mbd.filter_data(
        result_df,
        df_sirius_training,
        df_guid,
        output_file_structures_validation,
        output_file_extracted_massbank,
        output_file_guid_dtxsid,
    )

# * Get true fingerprints for training and validation molecules

# read fingerprint definitions
sirius_fps_df_path = sirius_input_directory / SIRIUS_FPS_DEFINITIONS
df_csi = pd.read_csv(sirius_fps_df_path, sep="\t")

# TODO remove
if 0:
    for step in ["TRAINING", "VALIDATION"]:
        print(step)

        # set id for chemical, and input and output paths
        if step == "TRAINING":
            # id
            id = CHEM_ID

            # input
            fps_input_file = INVITRODB_CHEMICAL_LIST
            fps_input_path = invitro_input_directory / fps_input_file

            # output
            fps_output_path = fps_output_directory / TRUE_FPS_TRAINING

        elif step == "VALIDATION":
            # id
            id = MASSBANK_ID

            # input
            fps_input_file = EXTRACTED_MASSBANK_SUMMARY_SMILES
            fps_input_path = massbank_output_directory / fps_input_file

            # output
            fps_output_path = fps_output_directory / TRUE_FPS_VALIDATION

        else:
            print("Please specify a valid step.")

        # set sdf and csv output path
        base_file_name = os.path.splitext(fps_input_file)[0]
        sdf_output_path = os.path.join(
            intermediate_fps_output_directory, f"{base_file_name}_clean.sdf"
        )
        csv_output_path = os.path.join(
            intermediate_fps_output_directory, f"{base_file_name}_clean.csv"
        )

        # process molecules
        get_truefps.process_molecules(
            id,
            fps_input_path,
            df_csi,
            fps_output_path,
            csv_output_path,
            sdf_output_path,
        )


# * Process validation and application output from SIRIUS

# set input and output paths
for step in ["VALIDATION", "APPLICATION"]:
    if step == "VALIDATION":
        # input
        sirius_input_folder = (
            sirius_input_directory / SIRIUS_MASSBANK_VALIDATION_DIR_FOLDER
        )
        # complete massbank summary
        massbank_validation_input_file = (
            massbank_output_directory / EXTRACTED_MASSBANK_SUMMARY_COMPLETE
        )

        # output
        predfps_output_file = PRED_FPS_VALIDATION

    elif step == "APPLICATION":
        # input
        sirius_input_folder = sirius_input_directory / SIRIUS_APPLICATION_DIR_FOLDER
        # no massbank validation
        massbank_validation_input_file = None

        # output
        predfps_output_file = PRED_FPS_APPLICATION

    else:
        print("Please specify a valid step.")

    print("SIRIUS input folder (fpt files):", sirius_input_folder)

    # load true training fingerprints
    training_fps_input_path = fps_output_directory / TRUE_FPS_TRAINING
    df_truefps = ml.get_fingerprint_df(training_fps_input_path)

    # set path for output file
    predfps_output_path = fps_output_directory / predfps_output_file

    # check SIRIUS version
    matching_sirius_version = get_predfps.check_sirius_version(
        sirius_input_folder / SIRIUS_FPS_DEFINITIONS
    )
    if not matching_sirius_version:
        "Error: This data cannot be processed."
        sys.exit()

    # read SIRIUS .fpt files
    df_predfps = get_predfps.read_fps_files(
        sirius_input_folder,
        df_csi,
        threshold=0.5,
    )
    print(df_predfps.shape)
    print(df_predfps.columns)
    print(df_predfps)

    # only keep fingerprints bits used for training
    df_predfps_selected = get_predfps.keep_selected_fps_bits(
        df_predfps,
        df_truefps,
    )
    print(df_predfps_selected.shape)
    print(df_predfps_selected.columns)
    print(df_predfps_selected)

    # apply filtering and grouping (only relevant for MassBank validation data)
    if step == "VALIDATION":
        df_predfps_filtered = get_predfps.filter_mbval_data_by_accession(
            df_predfps_selected,
            massbank_validation_input_file,
        )
        # aggregate MassBank validation data
        df_predfps_grouped = df_predfps_filtered.groupby(MASSBANK_ID).agg(
            get_predfps.custom_agg
        )
        df_predfps_grouped["ACCESSION"] = df_predfps_grouped["ACCESSION"].apply(
            lambda x: sorted(x)
        )
        # drop columns
        columns_to_drop = [
            "sirius_id",
            "sirius_run",
            "formula",
            "adduct",
            "adjusted_sirius_run",
        ]
        df_predfps_grouped = df_predfps_grouped.drop(columns=columns_to_drop)
        df_predfps_final = df_predfps_grouped.copy()
        df_predfps_final = df_predfps_final.sort_index().reset_index()
    elif step == "APPLICATION":
        # remove "features" column for application data from PARC
        # TODO check if needed?
        # if SIRIUS_APPLICATION_DIR_FOLDER in ["sources_fractions"]:
        # df_predfps_selected = df_predfps_selected.drop(columns="features")
        df_predfps_final = df_predfps_selected.copy()

    print(
        f"The shape of the final data frame for step {step} with predicted fingerprints is: {df_predfps_final.shape}"
    )

    # store output
    # TODO sort by "features" column?
    df_predfps_final.to_parquet(predfps_output_path, index=False)


# * Perform SIRIUS validation
# TODO decide what we need for paper (only for supplementary)
# TODO only write code for that

# TODO filter for safe SIRIUS compounds (ie those not used for training SIRIUS)
# TODO 1 figure predicted vs true, false predictions marked in colors
# TODO report Tanimoto, recall, precision


# Load MassBank validation dataframes
input_true_fps_validation = fps_output_directory / TRUE_FPS_VALIDATION
input_pred_fps_validation = fps_output_directory / PRED_FPS_VALIDATION
df_mbval_true_fps = validate.load_massbank_val_df(input_true_fps_validation)
df_mbval_pred_fps = validate.load_massbank_val_df(input_pred_fps_validation)
# introduced after switching from csv to parquet files for fingerprints
if "sirius_id" in df_mbval_pred_fps.columns:
    df_mbval_pred_fps["sirius_id"] = df_mbval_pred_fps["sirius_id"].astype("int")
df_mbval_pred_fps = df_mbval_pred_fps.drop(
    columns=["features", "TopCSIScore", "ACCESSION"]
)

# Filter dataframes to only contain shared chemicals (GUIDs)
(df_mbval_true_fps_filtered, df_mbval_pred_fps_filtered) = validate.filter_mbval_dfs(
    df_mbval_true_fps, df_mbval_pred_fps
)

# Calculate validation metrics per chemical (GUID)
output_metrics = (
    validation_fps_output_directory / "massbank_validation_metrics_guid.csv"
)
df_metrics_chem = validate.calculate_metrics_per_chemical(
    df_mbval_true_fps_filtered, df_mbval_pred_fps_filtered, output_metrics
)

# Create discrepancy figure
output_figure = (
    validation_fps_output_directory / "massbank_validation_difference_heatmap.png"
)
validate.generate_discrepancy_figure(
    df_mbval_true_fps_filtered, df_mbval_pred_fps_filtered, output_figure
)

# Calculate metrics per fingerprint bit
output_metrics_bit = (
    validation_fps_output_directory / "massbank_validation_metrics_absindex.csv"
)
df_metrics_bit = validate.calculate_metrics_per_bit(
    df_mbval_true_fps_filtered, df_mbval_pred_fps_filtered, output_metrics_bit
)

# Create histogram
output_figure = (
    validation_fps_output_directory / "massbank_validation_metrics_histogram.png"
)
validate.plot_metrics_histogram(df_metrics_bit, output_figure)

# Create density plot
output_figure = (
    validation_fps_output_directory / "massbank_validation_metrics_density.png"
)
validate.plot_metrics_density(df_metrics_bit, output_figure)

# Only keep fingerprint bits which were fit well enough
threshold_recall = 0.75
threshold_precision = 0.75
output_file = validation_fps_output_directory / VALIDATION_SELECTED_ABSINDEX
validate.save_indexes_passed_validation(
    df_metrics_bit, threshold_recall, threshold_precision, output_file
)

# %%

# implement safe and unsafe compounds validation

# only keep chemicals which are also in predicted fps dataframe
# TODO why is this filtering step different from filtering in filter_mb_vals()?
df_mbval_true_fps_filtered2 = df_mbval_true_fps[
    df_mbval_true_fps[MASSBANK_ID].isin(df_mbval_pred_fps[MASSBANK_ID])
]
print(df_mbval_pred_fps.shape)
print(df_mbval_true_fps_filtered2.shape)

# Check if the MASSBANK_ID columns are the same in both dataframes
columns_are_same = (
    df_mbval_true_fps_filtered2[MASSBANK_ID].sort_values().reset_index(drop=True)
    == df_mbval_pred_fps[MASSBANK_ID].sort_values().reset_index(drop=True)
).all()
print(f"MASSBANK_ID columns are the same: {columns_are_same}")
duplicates_true_fps = (
    df_mbval_true_fps_filtered2.set_index(MASSBANK_ID).duplicated().sum()
)
# duplicates_pred_fps = df_mbval_pred_fps.set_index(MASSBANK_ID).duplicated().sum()
print(f"Number of duplicate rows in df_mbval_true_fps_filtered: {duplicates_true_fps}")
# print(f"Number of duplicate rows in df_mbval_pred_fps: {duplicates_pred_fps}")

# filter the columns of df_mbval_pred_fps to keep only the common columns
# this is not needed later on
list_cols = [c for c in df_mbval_true_fps.columns if c in df_mbval_pred_fps.columns]
df_mbval_pred_fps_filtered = df_mbval_pred_fps[list_cols]
print(df_mbval_pred_fps_filtered.shape)
print(df_mbval_pred_fps_filtered.columns)

# read the GUID_DTXSID massbank summary file to filter on
input_GUID_DTXSID = massbank_output_directory / EXTRACTED_MASSBANK_GUID_DTXSID
df_guid_dtxsid = pd.read_csv(input_GUID_DTXSID)

# merge the pred fingerprints with DTXSID from the massbank summary file
# (this adds a few entries as some GUIDs have several DTXSIDs)
df_mbval_pred_fps_relevant = pd.merge(
    df_mbval_pred_fps,
    df_guid_dtxsid,
    left_on=MASSBANK_ID,
    right_on=MASSBANK_ID,
    how="inner",
)

# sort columns
print(df_mbval_pred_fps_relevant)
list_cols = sorted(df_mbval_pred_fps_relevant.columns)
list_cols = [CHEM_ID, MASSBANK_ID] + [
    c for c in list_cols if c not in [CHEM_ID, MASSBANK_ID]
]
df_mbval_pred_fps_relevant = df_mbval_pred_fps_relevant[list_cols]

# calculate duplicated DTXSIDs
n_duplicated_dtxsid = df_mbval_pred_fps_relevant.duplicated(subset=[CHEM_ID]).sum()
print(f"Number of duplicated DTXSIDs: {n_duplicated_dtxsid}")

# for each duplicated DTXSID, we set a bit to 1 if it is active for one of the duplicates
aggregated_df = df_mbval_pred_fps_relevant.groupby(CHEM_ID).agg("max")

# check how many bits are affected
# list_cols = [c for c in df_mbval_pred_fps_relevant.columns if c.isdigit()]
# list_cols = [CHEM_ID] + list_cols
# df_agg = df_mbval_pred_fps_relevant[list_cols].groupby(CHEM_ID).agg("mean")
# (df_agg.isin([0, 1]).sum(axis=1) == 1797).sum()

# %%

if MASSBANK_ID in aggregated_df.columns:
    aggregated_df = aggregated_df.drop(columns=MASSBANK_ID)
aggregated_df = aggregated_df.reset_index()

print(aggregated_df.shape)
print(aggregated_df.columns)

# print to file to be used as reference for ML
output_file = massbank_output_directory / VALIDATION_PRED_FPS_DTXSID
aggregated_df.sort_values(CHEM_ID).to_csv(output_file, index=False)
# %%



# Data processing: 2) process molecules
import re

import warnings

import pandas as pd
import numpy as np
from tqdm import tqdm
import logging

from rdkit import Chem
from rdkit.Chem.MolStandardize import rdMolStandardize
from openbabel import pybel
from CDK_pywrapper import CDK, FPType

from mlinvitrotox.constants import (
    # IDs
    SMILES_ID,
)


def remove_illegal_smiles(df, smiles_column, nonallowed):
    """Remove rows where SMILES string contains specific non-allowed strings or characters"""
    nonallowed_pattern = "|".join(map(re.escape, nonallowed))
    mask = df[smiles_column].str.contains(nonallowed_pattern, na=False)
    df_clean = df[~mask]
    return df_clean


def standardize_smiles(raw_smiles):
    if raw_smiles is None:
        return None
    try:
        # Standardize and desalt the molecule
        std1_smiles = rdMolStandardize.StandardizeSmiles(raw_smiles)
        desalter = rdMolStandardize.LargestFragmentChooser()
        desalt_mol = desalter.choose(Chem.MolFromSmiles(std1_smiles))
        std2_smiles = rdMolStandardize.StandardizeSmiles(Chem.MolToSmiles(desalt_mol))
        return std2_smiles
    except Exception as e:
        # Log the exception and return None
        #        smiles = Chem.MolToSmiles(mol) if mol else "Invalid mol"
        logging.warning(f"Error processing molecule: {raw_smiles}; Error: {e}")
        return None


def standardize_mol(mol):
    if mol is None:
        return None
    try:
        # Standardize and desalt the molecule
        std1_mol = rdMolStandardize.StandardizeSmiles().standardize(
            Chem.MolToSmiles(mol)
        )
        desalter = rdMolStandardize.LargestFragmentChooser()
        desalt_mol = desalter.choose(std1_mol)
        std2_mol = rdMolStandardize.StandardizeSmiles().standardize(
            Chem.MolToSmiles(desalt_mol)
        )
        return std2_mol
    except Exception as e:
        # Log the exception and return None
        smiles = Chem.MolToSmiles(mol) if mol else "Invalid mol"
        logging.warning(f"Error processing molecule: {smiles}; Error: {e}")
        return None


def process_molecules(
    id,
    fps_input_path,
    df_csi,
    fps_output_path,
    csv_output_path,
    sdf_output_path,
):
    # Set ID for SMILES
    smiles = SMILES_ID

    # Read the CSV data
    data = pd.read_csv(fps_input_path)
    print("The shape of the input data frame with structures:")
    print(data.shape)

    # Check for duplicates
    if data[id].duplicated().any():
        warnings.warn(
            f"There are duplicates in the file {fps_input_path} based on '{id}'"
        )

    # Only keep SMILES and chemical ID
    keep_columns = [id, smiles]
    df = data[keep_columns].copy()
    print("The shape of the modified data frame:")
    print(df.shape)
    print(df.head())

    # pre-cleanup
    df.loc[:, id] = df[id].str.replace(
        "https://comptox.epa.gov/dashboard/chemical/details/", ""
    )
    df = remove_illegal_smiles(df, smiles, ["Zn", "Pt", "<", ">", "R"])
    df = df[df[smiles].str.count("[Cc]") >= 2]
    df.loc[:, "sdf"] = df[smiles].apply(lambda x: Chem.MolFromSmiles(x) if x else None)
    df.dropna(subset=["sdf"], inplace=True)

    # standardize according to chemdbl procedure
    df["standardized_smiles"] = df[smiles].apply(standardize_smiles)
    # df["standardized_mol"] = df["sdf"].apply(standardize_mol)
    # df["standardized_smiles"] = df["standardized_mol"].apply(
    # lambda x: Chem.MolToSmiles(x) if x else None
    # )
    df["standardized_mol"] = df["standardized_smiles"].apply(
        lambda x: Chem.MolFromSmiles(x) if x else None
    )

    # Store sdf file
    writer = Chem.SDWriter(sdf_output_path)
    for index, row in df.sort_values(id).iterrows():
        mol = row["standardized_mol"]
        chem_id = row[id]
        if mol is not None:
            mol.SetProp(id, chem_id)
            writer.write(mol)
    writer.close()

    # Store csv file
    columns_to_save = [id, smiles, "standardized_smiles"]
    df[columns_to_save].sort_values(id).to_csv(csv_output_path, index=False)
    print("The shape of the cleaned-up data frame with smiles and mols:")
    print(df.shape)

    # generate FP3 and FP4 OB fingerprints via pybel
    allmols_with_dtxsid = []
    for mol in pybel.readfile("sdf", sdf_output_path):
        if mol is not None:
            # Try to retrieve the molecule_id, default to "Unknown" if not found
            try:
                dtxsid = (
                    mol.OBMol.GetData(id).GetValue()
                    if mol.OBMol.HasData(id)
                    else "Unknown"
                )
            except Exception:
                dtxsid = "Unknown"
            allmols_with_dtxsid.append((dtxsid, mol))

    # FP3
    fps_FP3 = []
    fp3_bits = []
    for item in tqdm(allmols_with_dtxsid, desc="Calculating FP3", unit="mol"):
        dtxsid, mol = item  # Correctly unpack the tuple
        fp = mol.calcfp(fptype="FP3")
        fps_FP3.append({id: dtxsid, "fp": fp})
        fp3_bits.append({id: dtxsid, "bits": fp.bits})

    binary_vectors_fp3 = []
    n_bits = 55

    for item in tqdm(fps_FP3, desc="Processing FP3", unit="mol"):
        fp_bits = item["fp"].bits
        dtxsid = item[id]
        bv = np.zeros(n_bits, dtype=int)
        if fp_bits:
            fp_indexes = [i - 1 for i in fp_bits]
            bv[fp_indexes] = 1
        # Append the bits directly rather than as a 'vector' key
        binary_vectors_fp3.append(
            {**{f"fp3_bit_{i+1}": bv[i] for i in range(n_bits)}, id: dtxsid}
        )

    df_binary_vectors_fp3 = pd.DataFrame(binary_vectors_fp3)
    df_binary_vectors_fp3 = df_binary_vectors_fp3.set_index(id)
    print(f"columns of fp3 df: {df_binary_vectors_fp3.columns}")
    print(f"index of fp3 df: {df_binary_vectors_fp3.index}")

    # FP4
    fps_FP4 = []
    fp4_bits = []
    for item in tqdm(allmols_with_dtxsid, desc="Calculating FP4", unit="mol"):
        dtxsid, mol = item  # Correctly unpack the tuple
        fp = mol.calcfp(fptype="FP4")
        fps_FP4.append({id: dtxsid, "fp": fp})
        fp4_bits.append({id: dtxsid, "bits": fp.bits})

    binary_vectors_fp4 = []
    n_bits = 307  # Adjust the number of bits for FP4

    for item in tqdm(fps_FP4, desc="Processing FP4", unit="mol"):
        fp_bits = item["fp"].bits
        dtxsid = item[id]
        bv = np.zeros(n_bits, dtype=int)
        if fp_bits:
            fp_indexes = [i - 1 for i in fp_bits]
            bv[fp_indexes] = 1
        # Append the bits directly rather than as a 'vector' key
        binary_vectors_fp4.append(
            {**{f"fp4_bit_{i+1}": bv[i] for i in range(n_bits)}, id: dtxsid}
        )

    df_binary_vectors_fp4 = pd.DataFrame(binary_vectors_fp4)
    df_binary_vectors_fp4 = df_binary_vectors_fp4.set_index(id)
    print(f"columns of fp4 df: {df_binary_vectors_fp4.columns}")
    print(f"index of fp4 df: {df_binary_vectors_fp4.index}")

    # concatenate FP3 and FP4
    df_obabel = pd.concat([df_binary_vectors_fp3, df_binary_vectors_fp4], axis=1)
    print("The shape of the obabel data frame:")
    print(df_obabel.shape)
    print("The index of the obabel data frame:")
    print(df_obabel.index)

    # create MACCS, PubChem and Klekota Roth fingerprints with CKW_pywrapper
    mols_list = []
    chem_ids_list = []
    sdf_supplier = Chem.SDMolSupplier(sdf_output_path)

    for mol in tqdm(sdf_supplier, desc="Processing Molecules", unit="mol"):
        if mol is not None:
            mols_list.append(Chem.AddHs(mol))
            chem_ids_list.append(mol.GetProp(id))

    # calculate fingerprints
    cdk = CDK(fingerprint=FPType.MACCSFP)
    df_maccs = cdk.calculate(mols_list, show_banner=False)
    cdk = CDK(fingerprint=FPType.PubchemFP)
    df_pubchem = cdk.calculate(mols_list, show_banner=False)
    cdk = CDK(fingerprint=FPType.KRFP)
    df_klekotaroth = cdk.calculate(mols_list, show_banner=False)

    # concatenate CDK fingerprints
    df_cdk = pd.concat((df_maccs, df_pubchem, df_klekotaroth), axis=1)
    df_cdk.index = chem_ids_list
    print("The shape of the CDK data frame:")
    print(df_cdk.shape)

    # concatenate CKD and pybel fingerprints
    df_fps = pd.concat((df_obabel, df_cdk), axis=1)
    df_fps.dropna(inplace=True)
    print("The shape of the initial combined obabel and pyfps data frame:")
    print(df_fps.shape)
    df_fps.columns = range(len(df_fps.columns))

    # CSI FingerID definitions
    df_csi["absoluteIndex"] = df_csi["absoluteIndex"].astype(int)
    valid_indices = set(df_csi["absoluteIndex"]).intersection(set(df_fps.columns))
    filtered_df_fps = df_fps[list(valid_indices)]
    filtered_df_fps.columns = [str(col).zfill(4) for col in filtered_df_fps.columns]

    # return filtered_df_fps

    # Define the ranges of each fingerprint type
    # This ranges are only informative and not used later on.
    openbabel_fp3_range = range(0, 55)
    openbabel_fp4_range = range(55, 362)
    MACCS_range = range(369, 525)
    PubChem_range = range(659, 1395)
    klekota_range = range(1409, 6267)
    print_info = (
        "The absoluteIndex from csi_fingerid.tsv file "
        "are used as column names. Following fingerprints are generated:\n"
        f"openbabel_fp3 fingerprints (absoluteIndex range {openbabel_fp3_range.start} - {openbabel_fp3_range.stop - 1}),\n"
        f"openbabel_fp4 (CDK Substructure in SIRIUS, absoluteIndex range {openbabel_fp4_range.start} - {openbabel_fp4_range.stop - 1}),\n"
        f"MACCS (absoluteIndex range {MACCS_range.start} - {MACCS_range.stop - 1}),\n"
        f"PubChem (absoluteIndex range {PubChem_range.start} - {PubChem_range.stop - 1}),\n"
        f"Klekota-Roth (absoluteIndex range {klekota_range.start} - {klekota_range.stop - 1})."
    )
    print(print_info)

    filtered_df_fps = filtered_df_fps.reset_index().rename(columns={"index": id})
    filtered_df_fps = filtered_df_fps.sort_values(id)
    filtered_df_fps.to_parquet(fps_output_path, index=False)
    # filtered_df_fps.to_csv(fps_output_path, index=False)

    print("The shape of the final data frame with fingerprints:")
    print(filtered_df_fps.shape)



import joblib
import zipfile
import shutil
import requests
import toml
from io import StringIO

from pathlib import Path

import pandas as pd

import click
from tqdm import tqdm

import mlinvitrotox.utils.training as ml
import mlinvitrotox.utils.get_predicted_fps as get_predfps
import mlinvitrotox.utils.predict as predict

SELECTED_AEIDS = "selected_aeids.csv"
SELECTED_CHEMICALS = "selected_chemicals.csv"


class Model:
    model_repo = "https://zenodo.org/records/13323297/files/"

    def __init__(
        self, target_variable: str = "hitcall_c", ml_algorithm: str = "classification"
    ) -> None:
        self.target_variable = target_variable
        self.ml_algorithm = ml_algorithm

    def load(self, path: str | Path):
        model_path = Path(path)
        self.model_path = model_path

        # if it doesn't exist, check if it's a model name that we can download from zenodo
        if not model_path.exists():
            try:
                model_name = model_path.with_suffix("")
                model_path = self.download(model_name)
            except RuntimeError as e:
                raise FileNotFoundError(
                    f"Could not find the model file {model_path}. To use the provided models, specify '-m mlinvitrotox_model'."
                )

        if model_path.suffix == ".itox":
            # Open the zip file
            zip_ref = zipfile.ZipFile(model_path, "r")

            def get_path(sub_path):
                # Open the specific CSV file inside the zip
                try:
                    return zip_ref.open(Path(sub_path).as_posix())
                except KeyError as e:
                    raise FileNotFoundError(e)
        elif model_path.is_dir():

            def get_path(sub_path):
                return model_path / sub_path
        else:
            raise RuntimeError(f"Failed to open file {model_path}")

        self.get_path = get_path

        # load true training fingerprints
        training_fps_path = get_path("training_true-fps.parquet")
        self.df_truefps = ml.get_fingerprint_df(training_fps_path)

        # load logged assay information overview
        assay_info_path = get_path(SELECTED_AEIDS)
        df_assay_info = pd.read_csv(assay_info_path)
        df_assay_info["aeid"] = df_assay_info["aeid"].astype("str")
        self.df_assay_info = df_assay_info

    def download(self, model_name: str):
        url = self.model_repo + "/" + str(model_name) + ".itox"
        appdir = Path(click.get_app_dir("mlinvitrotox"))

        appdir.mkdir(parents=True, exist_ok=True)

        # Define the path to the TOML file
        toml_file = Path(appdir) / "config.toml"

        # Default value for "model_cache"
        default_value = Path(appdir) / "model_cache/"

        default_value.mkdir(parents=True, exist_ok=True)

        # Check if the file exists
        if toml_file.exists():
            # Load the existing TOML file
            with open(toml_file, "r") as file:
                config = toml.load(file)
                model_cache = config.get("model_cache", str(default_value))
        else:
            # File does not exist, create it with default values
            config = {"model_cache": str(default_value)}
            with open(toml_file, "w") as file:
                toml.dump(config, file)
            model_cache = Path(default_value)

        local_filename = Path(model_cache) / f"{model_name}.itox"

        if local_filename.exists():
            return local_filename

        # Send a GET request to the URL
        response = requests.get(url, stream=True)

        if not response.ok:
            raise RuntimeError(
                f"Could not find the model {model_name} in the online model repository."
            )

        print("Downloading file ")
        # Open the local file in write-binary mode and save the contents
        with open(local_filename, "wb") as file:
            for chunk in tqdm(response.iter_content(chunk_size=8192)):
                file.write(chunk)

        print(f"Downloaded {local_filename} from {url}")

        return local_filename

    def predict(self, df_predfps):
        # only keep fingerprints bits used for training
        df_predfps_selected = get_predfps.keep_selected_fps_bits(
            df_predfps,
            self.df_truefps,
        )

        # initialize
        list_dfs_pred = []

        # for each aeid
        aeids = self.df_assay_info["aeid"].astype("str").to_list()

        for aeid in tqdm(aeids, desc="Processing AEIDs"):
            df_pred = self.predict_aeid(aeid, df_predfps_selected)
            if df_pred is not None:
                list_dfs_pred.append(df_pred)

        # concatenate
        df_predictions = pd.concat(list_dfs_pred)
        df_predictions["prediction"] = df_predictions["prediction"].astype(int)

        # merge predictions with assay information
        df_predictions = pd.merge(
            df_predictions,
            self.df_assay_info[["aeid", "MechanisticTarget", "signal_direction"]],
            on="aeid",
            how="left",
        )

        # Calculate score per endpoint
        df_predictions = predict.calculate_endpoint_score(df_predictions)

        return df_predictions

    def predict_aeid(self, aeid, df_predfps_selected):
        chem_ids = df_predfps_selected.index

        aeid_path = Path(self.target_variable) / self.ml_algorithm / str(aeid)
        if True:  # TODO add a check if file exists
            # get true fingerprints for selected chemicals
            selected_chemicals_file = self.get_path(aeid_path / SELECTED_CHEMICALS)
            df_truefps_selected = predict.get_true_fingerprints_for_selected_chemicals(
                selected_chemicals_file,
                self.df_truefps,
                self.target_variable,
            )

            # load classifier and preprocessing pipeline
            preproc_path = aeid_path / "Feature_Selection_RandomForestClassifier"
            try:
                preprocessor_path = self.get_path(
                    Path(preproc_path) / "preprocessing_model.joblib"
                )
                preprocessor = joblib.load(preprocessor_path)
            except FileNotFoundError:
                preprocessor = None

            try:
                classifier_path = self.get_path(
                    Path(preproc_path)
                    / "XGBClassifier"
                    / "best_estimator_full_data.joblib"
                )
                classifier = joblib.load(classifier_path)
            except FileNotFoundError:
                classifier = None

            # catch unsuccessful runs
            if classifier is None:
                return None
            # preprocess fingerprints
            preprocessed_truefps = preprocessor.transform(df_truefps_selected)
            preprocessed_predfps = preprocessor.transform(df_predfps_selected)

            # calculate maximum cosine similarity
            max_similarities = predict.calculate_maximum_similarity(
                preprocessed_predfps,
                preprocessed_truefps,
            )

            # calculate predictions for each data point
            df_pred = predict.calculate_predictions(
                classifier,
                preprocessed_predfps,
                chem_ids,
                max_similarities,
                aeid,
            )
            return df_pred

    def export(self, filepath: str | Path, export_stats=False):
        """Exports current model to an .itox file at the given path

        Args:
            filepath (str | Path): the path to the .itox file to be created
            export_stats (boolean): if the statistics files should be exported as well

        """
        filepath = Path(filepath)
        if filepath.suffix != ".itox" and not filepath.is_dir():
            raise RuntimeError(
                f"Wrong file type of target. Export targets can only be .itox file or an existing directory.\nProvided filepath is {filepath}"
            )

        if self.model_path.suffix == ".itox":
            # since we already have an (immutable) itox file we can just copy it
            shutil.copy(
                self.model_path,
                filepath,
            )
        elif self.model_path.is_dir():
            # We need to collect all relevant files for export
            if filepath.is_dir():
                # make the output have the same filename as the original
                filepath = filepath / self.model_path.with_suffix(".itox").name

            include_list = []

            # add base
            include_list += [self.model_path / "modeltraining_metrics.csv"]
            include_list += [self.model_path / "training_true-fps.parquet"]
            # selected_aeids.csv will be created later

            # add logs
            include_list += [self.model_path / "log" / "modelfitting.log"]

            # for each aeid
            failed_aeids = []
            aeid_folders = self.model_path / self.target_variable / self.ml_algorithm
            for aeid_folder in aeid_folders.glob("*/"):
                if (
                    aeid_folder / "Feature_Selection_RandomForestClassifier"
                    not in aeid_folder.iterdir()
                ):
                    continue
                if aeid_folder / "failed.txt" in aeid_folder.iterdir():
                    failed_aeids += [int(aeid_folder.stem)]
                    continue

                include_list += [aeid_folder / "selected_chemicals.csv"]
                feature_selection_path = (
                    aeid_folder / "Feature_Selection_RandomForestClassifier"
                )
                include_list += [feature_selection_path / "preprocessing_model.joblib"]
                include_list += [feature_selection_path / "selected_features.csv"]
                estimator_path = (
                    aeid_folder
                    / "Feature_Selection_RandomForestClassifier"
                    / "XGBClassifier"
                )
                include_list += [estimator_path / "best_estimator_full_data.joblib"]

                # add metric folders
                if export_stats:
                    include_list += list(estimator_path.rglob("mb_val_sirius/*"))
                    include_list += list(estimator_path.rglob("mb_val_structure/*"))
                    include_list += list(estimator_path.rglob("test/*"))

            # update selected aeids files to not contain failed aeids
            selected_aeids_path = self.get_path("selected_aeids.csv")
            df = pd.read_csv(selected_aeids_path)
            df = df[~df["aeid"].isin(failed_aeids)]

            # Writing the files to the zipfile
            with zipfile.ZipFile(
                filepath, "w", compression=zipfile.ZIP_DEFLATED
            ) as zipf:
                for path in tqdm(include_list):
                    zipf.write(path, arcname=path.relative_to(self.model_path))

                # write the selected_aeids.csv
                csv_buffer = StringIO()
                df.to_csv(csv_buffer, index=False)
                zipf.writestr("selected_aeids.csv", csv_buffer.getvalue())
        else:
            raise RuntimeError(f"Unkown model file {self.model_path}")



import os

from pathlib import Path

import numpy as np
import pandas as pd

import plotly.express as px

import streamlit as st

import mlinvitrotox.app.helper as app

from mlinvitrotox.constants import (
    # Paths
    # MODELS_RESULTS_DIR_PATH,
    # RESULTS_DIR_PATH,
    # Files
    SELECTED_AEIDS,
)

# reverse renaming dictionary
dict_reverse_rename = {v: k for k, v in app.dict_rename.items()}

# Add a dropdon to the sidebar: Select run folder
logs_folder = Path("models")
# logs_folder = Path(MODELS_RESULTS_DIR_PATH)
run_folder = st.sidebar.selectbox(
    "Select Run", [run_folder for run_folder in os.listdir(logs_folder)][::-1]
)
run_folder_path = logs_folder / run_folder

# Load logged assay information overview
assay_info_path = run_folder_path / SELECTED_AEIDS
assay_info_df = pd.read_csv(assay_info_path)
aeid_list = [str(i) for i in sorted(assay_info_df["aeid"].to_list())]
assay_info_df["aeid"] = assay_info_df["aeid"].astype("str")

# Load metrics
metrics_path = run_folder_path / "modeltraining_metrics.csv"
# metrics_path = Path(RESULTS_DIR_PATH) / "predictions" / f"{run_folder}_metrics.csv"
metrics_df = pd.read_csv(metrics_path)
metrics_df["aeid"] = metrics_df["aeid"].astype("str")

# Add a dropdown to the sidebar: Select target variable
target_variable_list = os.listdir(run_folder_path)
target_variable_list = [
    d
    for d in target_variable_list
    if os.path.isdir(run_folder_path / d) and "hitcall" in d
]
if len(target_variable_list) > 1:
    target_variable = st.sidebar.selectbox(
        "Select Target Variable", target_variable_list
    )
else:
    target_variable = target_variable_list[0]
    st.sidebar.write(f"Target variable: {target_variable}")

# Add a dropdown to the sidebar: Select algorithm
target_variable_path = run_folder_path / target_variable
ml_algorithm_list = os.listdir(target_variable_path)
ml_algorithm_list = [
    d for d in ml_algorithm_list if os.path.isdir(target_variable_path / d)
]
if len(ml_algorithm_list) > 1:
    ml_algorithm = st.sidebar.selectbox(
        "Select ML Algorithm",
        ml_algorithm_list,
    )
else:
    ml_algorithm = ml_algorithm_list[0]
    st.sidebar.write(f"Algorithm: {ml_algorithm}")

# Select dummy AEID
# TODO how to select it well?
dummy_aeid = str(100)

# Add a dropdown to the sidebar: Select preprocessing model
aeid_path = target_variable_path / ml_algorithm / dummy_aeid
preprocessor_list = os.listdir(aeid_path)
preprocessor_list = [d for d in preprocessor_list if os.path.isdir(aeid_path / d)]
if len(preprocessor_list) > 1:
    preprocessor = st.sidebar.selectbox(
        "Select feature selection model", preprocessor_list
    )
else:
    preprocessor = preprocessor_list[0]
    st.sidebar.write(f"Feature selection: {preprocessor.split('_')[2]}")

# Add a dropdown to the sidebar: Select validation type
# TODO unify names with other dashboard page
val_type = st.sidebar.selectbox(
    "Select validation set",
    [
        "Internal validation",
        "MB validation from structure",
        "MB validation SIRIUS-predicted",
    ],
)
reverse_val_type = dict_reverse_rename[val_type]

# Add a dropdown to the sidebar: Select classification threshold
thresholds = ["default=0.5", "cost(TPR,TNR)", "TPR≈0.5", "TNR≈0.5"]
threshold = st.sidebar.selectbox(
    "Select classification threshold",
    thresholds,
)
reverse_threshold = dict_reverse_rename[threshold]
# Add a dropdown to the sidebar: Select class metric
class_metrics = ["macro avg", "weighted avg", "positive", "negative", "accuracy"]
support_class_metrics = ["macro avg", "positive", "negative"]
class_metric = st.sidebar.selectbox(
    "Select Class Slice",  # TODO better name!
    class_metrics,
)
reverse_class_metric = dict_reverse_rename[class_metric]
st.sidebar.divider()

# Add selectboxes for x and y axes metrics
# TODO or have another system to choose x and y
metrics = [
    "Recall (macro)",
    "Precision (macro)",
    "F1 (macro)",
    "Recall (micro)",
    "Precision (micro)",
    "F1 (micro)",
    "Balanced Accuracy",
    "ROC AUC",
    "PR AUC",
]
x_metric = st.sidebar.selectbox(
    "Select metric for x-axis",
    metrics,
)
y_metric = st.sidebar.selectbox(
    "Select metric for y-axis",
    [m for m in metrics if m != x_metric],
)
st.sidebar.divider()

# Sidebar: Plotting details
marginal_type = st.sidebar.selectbox(
    "Select marginal type",
    ["box", "violin", "rug", "histogram", None],
)
marker_size_category = st.sidebar.selectbox(
    "Select marker size based on",
    [
        "Higher Imbalance = LARGER",
        "Total Support",
        "Support Positive",
        "Support Negative",
        None,
    ],
)

color_palette_mapping = {
    "Plotly": px.colors.qualitative.Plotly,
    "Safe": px.colors.qualitative.Safe,
    "Light24": px.colors.qualitative.Light24,
    "Bold": px.colors.qualitative.Bold,
    "D3": px.colors.qualitative.D3,
    "G10": px.colors.qualitative.G10,
}

marker_symbol = "circle-dot"
opacity = 1.0
marker_line_width = 0.8
marker_line_color = "#000000"
# TODO which color palette? --> Plotly
# TODO find a color palette for all mechanistic targets!
palette = st.sidebar.selectbox(
    "Select color palette",
    list(color_palette_mapping.keys()),
)
palette = color_palette_mapping.get(palette)
autoscale = st.sidebar.checkbox("Autoscale", value=True)


st.title("Summary of validation results")

# set some values
metrics_df["Target Variable"] = app.dict_rename[target_variable]
metrics_df["Feature Selection"] = app.dict_rename[preprocessor.split("_")[2]]
metrics_df["Validation Set"] = app.dict_rename[val_type]

# extract information
# TODO how to get estimator info to metrics_df?
# metrics_df["Estimator"] = metrics_df["Estimator"].apply(lambda x: app.dict_rename[x])
metrics_df["Estimator"] = "XGBoost"

metrics_df = metrics_df.rename(
    columns={
        "recall_micro": "Recall (micro)",
        "recall_macro": "Recall (macro)",
        "precision_micro": "Precision (micro)",
        "precision_macro": "Precision (macro)",
        "f1_micro": "F1 (micro)",
        "f1_macro": "F1 (macro)",
        "balanced_accuracy": "Balanced Accuracy",
        "mcc": "MCC",
        "roc_auc": "ROC AUC",
        "pr_auc": "PR AUC",
        "support": "Total Support",
        "n_pos": "Support Positive",
    }
)
metrics_df["Support Negative"] = (
    metrics_df["Total Support"] - metrics_df["Support Positive"]
)

metrics_df["Imbalance"] = (
    metrics_df["Support Negative"] - metrics_df["Support Positive"]
)
# TODO how is imbalance score calculated?
metrics_df["Imbalance Score"] = (
    metrics_df["Support Positive"] - metrics_df["Support Negative"]
) / metrics_df["Total Support"]
metrics_df["Imbalance Score"] = metrics_df["Imbalance Score"].round(3)

# merge with mechanistic target information
metrics_df = pd.merge(
    metrics_df,
    assay_info_df[["aeid", "assay_component_endpoint_name", "MechanisticTarget"]],
    on="aeid",
    how="left",
)

# Apply the acronym mapping to the DataFrame
metrics_df["MechanisticTargetAcronym"] = metrics_df["MechanisticTarget"].map(
    app.acronym_mapping
)

df = metrics_df.copy()

st.header("Summary figure")

# apply factor for each marker size category
if marker_size_category == "Total Support":
    df["Marker Size"] = df["Total Support"] ** 0.42
elif marker_size_category == "Support Positive":
    df["Marker Size"] = df["Support Positive"] ** 0.55
elif marker_size_category == "Support Negative":
    df["Marker Size"] = df["Support Negative"] ** 0.45
elif marker_size_category == "Higher Imbalance = LARGER":
    df["Marker Size"] = np.abs(df["Imbalance Score"]) * 15
else:
    df["Marker Size"] = 10
df["Marker Size"] = df["Marker Size"].round(3)
# TODO add legend for marker size

histogram_args = (
    {"marginal_x": marginal_type, "marginal_y": marginal_type}
    if marginal_type is not None
    else {}
)

# TODO which hover data?
args = {
    "opacity": opacity,
    "hover_data": [
        "aeid",
        "Validation Set",
        "Precision (micro)",
        "Precision (macro)",
        "Recall (micro)",
        "Recall (macro)",
        "F1 (micro)",
        "F1 (macro)",
        # "Accuracy",
        "Balanced Accuracy",
        "ROC AUC",
        "PR AUC",
        "Total Support",
        "Support Positive",
        "Support Negative",
        "Imbalance Score",
    ],
    "color_discrete_sequence": palette,
    "custom_data": ["Marker Size"],  # Add customdata to match marker size
}

# Define axis font size and margin
axis_font_size = 20.5
margin = 20

# Construct the title string using the selected metrics
title = ""

# Scatter plot
fig = px.scatter(
    df,
    x=x_metric,
    y=y_metric,
    color="MechanisticTargetAcronym",
    **args,
    **histogram_args,
)


# Update the marker size based on the customdata
for j, trace in enumerate(fig.data):
    if "scatter" in trace.type:
        # Get the marker size from the customdata
        marker_size = list(fig.data[j].customdata[:, 0])
        marker_line_color_ = "black" if marker_line_color else None
        marker_line_width_ = marker_line_width if marker_line_color else None
        fig.data[j].update(
            marker=dict(
                symbol=marker_symbol,
                size=marker_size,
                line=dict(color=marker_line_color_, width=marker_line_width_),
            )
        )

# Update axis labels based on selected metrics
fig.update_layout(
    title=title,
    title_font=dict(size=21, color="black"),
    xaxis_title=x_metric,
    yaxis_title=y_metric,
    xaxis_title_font=dict(size=axis_font_size, color="black"),
    yaxis_title_font=dict(size=axis_font_size, color="black"),
    margin=dict(t=margin),
    width=1200,  # Increase the figure width
    height=800,  # Increase the figure height
    legend=dict(
        orientation="h",  # Horizontal orientation
        yanchor="top",  # Anchor legend at the top of its bounding box
        y=-0.2,  # Position legend below the x-axis
        xanchor="center",  # Anchor legend in the center of its bounding box
        x=0.5,  # Center the legend horizontally
        title_text="",  # Remove legend title
    ),
)

if autoscale:
    fig.update_layout(autosize=True)
else:
    fig.update_layout(xaxis=dict(range=[0.0, 1.0]))
    fig.update_layout(yaxis=dict(range=[0.0, 1.0]))

st.plotly_chart(fig, use_container_width=True)


st.subheader("Performance metrics across the target assay endpoints")
metrics_df = metrics_df[
    [
        "aeid",
        "assay_component_endpoint_name",
        "MechanisticTarget",
        "MechanisticTargetAcronym",
        # "Target Variable",
        # "Feature Selection",
        "Total Support",
        "Support Positive",
        "Support Negative",
        "Precision (micro)",
        "Precision (macro)",
        "Recall (micro)",
        "Recall (macro)",
        "F1 (micro)",
        "F1 (macro)",
        "Balanced Accuracy",
        "MCC",
        "ROC AUC",
        "PR AUC",
        "Estimator",
    ]
]
metrics_df = metrics_df.rename(
    columns={
        "aeid": "aeid",
        "assay_component_endpoint_name": "aeid name",
        "MechanisticTarget": "Mech. target",
        "MechanisticTargetAcronym": "MT acronym",
        # "Target Variable": "y",
        # "Feature Selection": "f. s.",
        "Total Support": "Support",
        "Support Positive": "Pos. support",
        "Support Negative": "Neg. support",
        "Precision (micro)": "micro prec.",
        "Precision (macro)": "macro prec.",
        "Recall (micro)": "micro rec.",
        "Recall (macro)": "macro rec.",
        "F1 (micro)": "micro F1",
        "F1 (macro)": "macro F1",
        "Balanced Accuracy": "bacc",
        "MCC": "mcc",
        "ROC AUC": "roc-auc",
        "PR AUC": "pr-auc",
        "Estimator": "model",
    }
)

st.dataframe(metrics_df, use_container_width=True, hide_index=True)

# TODO combine entries per acronym as a list
mechtar_df = (
    df[["MechanisticTargetAcronym", "MechanisticTarget"]]
    .drop_duplicates()
    .sort_values("MechanisticTargetAcronym")
    .reset_index(drop=True)
)
st.markdown("### Mechanistic targets and their acronyms")
st.dataframe(mechtar_df, hide_index=True)


# TODO Why median? When we include aeid in the groupby, it does not actually do a groupby and
#      the median corresponds to the value of the aeid
# st.subheader("Median performance metrics across the target assay endpoints")

# numeric_columns = [
# "Precision (micro)",
# "Precision (macro)",
# "Recall (micro)",
# "Recall (macro)",
# "F1 (micro)",
# "F1 (macro)",
# "Balanced Accuracy",
# "MCC",
# "ROC AUC",
# "PR AUC",
# "Total Support",
# "Support Positive",
# "Support Negative",
# ]

# groupby_columns = [
# "Target Variable",
# "aeid",
# "assay_component_endpoint_name",
# "Feature Selection",
# "Estimator",
# "Validation Set",
# ]
# mechtar_columns = [
# "MechanisticTarget",
# "MechanisticTargetAcronym",
# ]
# grouped_numeric = (
# df[groupby_columns + numeric_columns]
# .groupby(groupby_columns)
# .median()
# .reset_index()
# )
# grouped = pd.merge(
# grouped_numeric,
# df[groupby_columns + mechtar_columns].drop_duplicates(),
# on=groupby_columns,
# )

# grouped["Precision (micro)"] = grouped["Precision (micro)"].apply(lambda x: f"{x:.3f}")
# grouped["Precision (macro)"] = grouped["Precision (macro)"].apply(lambda x: f"{x:.3f}")
# grouped["Recall (micro)"] = grouped["Recall (micro)"].apply(lambda x: f"{x:.3f}")
# grouped["Recall (macro)"] = grouped["Recall (macro)"].apply(lambda x: f"{x:.3f}")
# grouped["F1 (micro)"] = grouped["F1 (micro)"].apply(lambda x: f"{x:.3f}")
# grouped["F1 (macro)"] = grouped["F1 (macro)"].apply(lambda x: f"{x:.3f}")
# grouped["Balanced Accuracy"] = grouped["Balanced Accuracy"].apply(lambda x: f"{x:.3f}")
# grouped["MCC"] = grouped["MCC"].apply(lambda x: f"{x:.3f}")
# grouped["ROC AUC"] = grouped["ROC AUC"].apply(lambda x: f"{x:.3f}")
# grouped["PR AUC"] = grouped["PR AUC"].apply(lambda x: f"{x:.3f}")

# summary = grouped[
# [
# "aeid",
# "assay_component_endpoint_name",
# "MechanisticTarget",
# "MechanisticTargetAcronym",
##"Target Variable",
##"Feature Selection",
# "Total Support",
# "Support Positive",
# "Support Negative",
# "Precision (micro)",
# "Precision (macro)",
# "Recall (micro)",
# "Recall (macro)",
# "F1 (micro)",
# "F1 (macro)",
# "Balanced Accuracy",
# "MCC",
# "ROC AUC",
# "PR AUC",
# "Estimator",
# ]
# ]
# summary = summary.rename(
# columns={
# "aeid": "aeid",
# "assay_component_endpoint_name": "aeid name",
# "MechanisticTarget": "Mech. target",
# "MechanisticTargetAcronym": "MT acronym",
##"Target Variable": "y",
##"Feature Selection": "f. s.",
# "Total Support": "Support",
# "Support Positive": "Pos. support",
# "Support Negative": "Neg. support",
# "Precision (micro)": "micro prec.",
# "Precision (macro)": "macro prec.",
# "Recall (micro)": "micro rec.",
# "Recall (macro)": "macro rec.",
# "F1 (micro)": "micro F1",
# "F1 (macro)": "macro F1",
# "Balanced Accuracy": "bacc",
# "MCC": "mcc",
# "ROC AUC": "roc-auc",
# "PR AUC": "pr-auc",
# "Estimator": "model",
# }
# )


# summary["aeid"] = summary["aeid"].astype("int")
# summary = summary.sort_values("aeid")
# st.dataframe(summary, use_container_width=True, hide_index=True)



import click
from streamlit import config as _config
from streamlit.web.bootstrap import run


@click.command("app")
def start_app():
    """
    Start the app

    """

    _config.set_option("server.headless", True)
    run("./src/mlinvitrotox/app/Results.py", args=[], flag_options={}, is_hello=False)


if __name__ == "__main__":
    start_app()



import os
import joblib

from pathlib import Path

import pandas as pd
import numpy as np

from sklearn.metrics.pairwise import cosine_similarity

from mlinvitrotox.constants import (
    # IDs
    CHEM_ID,
)


def get_aeids_from_path(base_path):
    """
    Get a list of AEIDs from the base_path directory.

    Parameters:
    - base_path: The base directory containing AEID directories.

    Returns:
    A list of AEIDs (as strings) found in the base_path.
    """
    # List all items in base_path
    items = os.listdir(base_path)
    # Filter items to only include directories (assuming AEIDs are directories)
    aeids = [item for item in items if os.path.isdir(Path(base_path) / item)]
    return aeids


def load_application_data(
    fps_input_file,
):
    if isinstance(fps_input_file, Path) and fps_input_file.suffix == ".csv":
        df = pd.read_csv(fps_input_file)
    else:
        df = pd.read_parquet(fps_input_file)

    # List of possible chemical ID columns
    possible_id_columns = ["dtxsid", "guid", "features", "sirius_id"]

    # Initialize a variable to store the name of the found ID column
    found_id_column = None

    # Check and rename the ID column
    cols_df = [col.lower() for col in df.columns]
    for column in possible_id_columns:
        if column.lower() in cols_df:
            found_id_column = column  # Store the found ID column name
            df.rename(columns={column: "chem_id"}, inplace=True)
            break  # Assuming only one such column exists, exit the loop after renaming

    # If no predefined ID column is found, prompt the user for selection
    if found_id_column is None:
        print(
            "No predefined ID column found. Please select the ID column from the list below:"
        )
        for index, column in enumerate(df.columns, start=1):
            print(f"{index}. {column}")
        selected_index = int(input("Enter the number of the ID column: ")) - 1
        if 0 <= selected_index < len(df.columns):
            selected_column = df.columns[selected_index]
            df.rename(columns={selected_column: "chem_id"}, inplace=True)
            found_id_column = selected_column
        else:
            print("Invalid selection. Exiting without renaming columns.")
            return df  # Early return in case of invalid selection

    # Filter columns to retain only the ID column and those with numbers in their names (considered feature columns)
    feature_columns = [
        col
        for col in df.columns
        if any(char.isdigit() for char in col) or col == "chem_id"
    ]
    df = df[feature_columns]
    return df


def get_true_fingerprints_for_selected_chemicals(
    selected_chemicals_file,
    truefps_df,
    target_variable,
):
    # load selected chemicals
    chem_df = pd.read_csv(selected_chemicals_file)

    # get true fingerprints for selected chemicals
    selected_truefps_df = truefps_df[
        truefps_df[CHEM_ID].isin(chem_df[CHEM_ID])
    ].reset_index(drop=True)
    columns_to_drop = [CHEM_ID, target_variable]
    selected_truefps_df = selected_truefps_df.drop(
        columns=[col for col in columns_to_drop if col in selected_truefps_df.columns]
    )

    return selected_truefps_df


def load_preprocessing_and_classifier(
    base_path,
):
    """
    Load a single classifier based on the provided path.

    Parameters:
    - base_path: The path

    Returns:
    The loaded classifier and preprocessing pipeline.
    """
    preprocessor_path = Path(base_path) / "preprocessing_model.joblib"
    classifier_path = (
        Path(base_path) / "XGBClassifier" / "best_estimator_full_data.joblib"
    )

    try:
        preprocessing_model = joblib.load(preprocessor_path)
    except FileNotFoundError:
        preprocessing_model = None

    try:
        classifier = joblib.load(classifier_path)
    except FileNotFoundError:
        classifier = None

    return classifier, preprocessing_model


def calculate_maximum_similarity(
    preprocessed_predfps,
    preprocessed_truefps,
):
    # calculate cosine similarity for each pair of chemicals
    cos_sim_matrix = cosine_similarity(
        preprocessed_predfps,
        preprocessed_truefps,
    )
    # take the maximum similarity for each application chemical to any of the training chemicals
    max_similarities = cos_sim_matrix.max(axis=1)

    return max_similarities


def calculate_predictions(
    classifier,
    preprocessed_predfps,
    chem_ids,
    max_similarities,
    aeid,
):
    # get predictions
    predictions = classifier.predict(preprocessed_predfps)
    prediction_probs = classifier.predict_proba(preprocessed_predfps)[:, 1]

    # combine with other features
    pred_df = pd.concat(
        (
            pd.Series(chem_ids, name="chem_id"),
            pd.Series(predictions, name="prediction"),
            pd.Series(prediction_probs, name="probability"),
            pd.Series(max_similarities, name="similarity"),
        ),
        axis=1,
    )
    pred_df.insert(loc=0, column="aeid", value=aeid)

    return pred_df


def calculate_endpoint_score(predictions_df):
    # remove not annotated mechanistic targets
    filtered_predictions_df = predictions_df[
        predictions_df["MechanisticTarget"] != "not_annotated"
    ]

    # group and calculate endpoint score
    cols_gb = ["chem_id", "MechanisticTarget", "signal_direction"]
    endpoint_score = (
        filtered_predictions_df.groupby(cols_gb)["prediction"]
        .agg(lambda x: x.sum() / len(x))
        .reset_index(name="endpoint_score")
    )
    # merge with predictions data frame
    predictions_df = pd.merge(
        predictions_df,
        endpoint_score,
        on=cols_gb,
        how="left",
    )
    predictions_df["endpoint_score"] = predictions_df["endpoint_score"].fillna(np.nan)

    # calculate number of endpoints per mechanistic target
    endpoints_per_mech_target = (
        filtered_predictions_df.groupby(cols_gb[1:])["aeid"]
        .nunique()
        .reset_index(name="endpoints_count")
    )
    # merge with predictions data frame
    predictions_df = pd.merge(
        predictions_df,
        endpoints_per_mech_target,
        on=cols_gb[1:],
        how="left",
    )

    return predictions_df


def sort_by_aeid_and_chem_id(df):
    df["sort"] = df["chem_id"].str.split("_", expand=True)[0].astype("int")
    df["aeid"] = df["aeid"].astype("int")
    df = df.sort_values(["aeid", "sort"])
    df = df.drop(columns=["sort"])
    return df


def sort_by_column(df, col):
    df["sort"] = df[col].str.split("_", expand=True)[0].astype("int")
    df = df.sort_values(["sort"])
    df = df.drop(columns=["sort"])
    return df



import os

# Root and Base Directories
ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
CONFIG_DIR_PATH = os.path.join(ROOT_DIR, "config")
DATA_DIR_PATH = os.path.join(ROOT_DIR, "../../data")
INPUT_DIR_PATH = os.path.join(DATA_DIR_PATH, "input")
OUTPUT_DIR_PATH = os.path.join(DATA_DIR_PATH, "processed")
RESULTS_DIR_PATH = os.path.join(DATA_DIR_PATH, "results")

# IDs
CHEM_ID = "DTXSID"
MASSBANK_ID = "GUID"
SMILES_ID = "SMILES"
SIRIUS_ID = "sirius_id"

# Data processing

# Step 1 input
MASSBANK_INPUT_DIR_PATH = os.path.join(INPUT_DIR_PATH, "hrms_massbank")
MASSBANK_INPUT_DIR_FOLDER = "published"
MASSBANK_UNIQUE_GUID = "massbank_unique_guids.csv"
SIRIUS_INPUT_DIR_PATH = os.path.join(INPUT_DIR_PATH, "sirius")
SIRIUS_TRAINING_STRUCTURES_INCHI = "sirius_training_structures_positive-mode_inchi.tsv"
SIRIUS_TRAINING_STRUCTURES_INCHI_DTXSID = (
    "sirius_training_structures_positive-mode_inchi_dtxsid.csv"
)

# Step 1 output
MASSBANK_OUTPUT_DIR_PATH = os.path.join(OUTPUT_DIR_PATH, "hrms_massbank")
EXTRACTED_MASSBANK_SUMMARY_SMILES = "massbank_published-data_smiles.csv"
EXTRACTED_MASSBANK_SUMMARY_COMPLETE = "massbank_published-data_complete.csv"
EXTRACTED_MASSBANK_GUID_DTXSID = "massbank_GUID-DTXSID.csv"


# Step 2 input
INVITRO_INPUT_DIR_PATH = os.path.join(INPUT_DIR_PATH, "invitro")
INVITRODB_CHEMICAL_LIST = "Chemical_List_ToxCast_invitroDB_v4_1-2024-04-29.csv"
SIRIUS_FPS_DEFINITIONS = "csi_fingerid.tsv"

# Step 2 output
FPS_OUTPUT_DIR_PATH = os.path.join(OUTPUT_DIR_PATH, "chemical_fingerprints")
INTERMEDIATE_FPS_OUTPUT_DIR_PATH = os.path.join(FPS_OUTPUT_DIR_PATH, "intermediate")
TRUE_FPS_TRAINING = "training_true-fps.parquet"
TRUE_FPS_VALIDATION = "validation_true-fps.parquet"


# Step 3 input
SIRIUS_MASSBANK_VALIDATION_DIR_FOLDER = "massbank_toxicity_5o8_eawag_sept2023_all"
SIRIUS_APPLICATION_DIR_FOLDER = "sources_fractions"

# Step 3 output
PRED_FPS_VALIDATION = "validation_sirius-pred-fps.parquet"
PRED_FPS_APPLICATION = "application_sirius-pred-fps.parquet"


# Step 4 input
# < no new input files >

# Step 4 output
VALIDATION_FPS_OUTPUT_DIR_PATH = os.path.join(FPS_OUTPUT_DIR_PATH, "validation")
VALIDATION_SELECTED_ABSINDEX = "massbank_validation_selected_absindex.csv"
VALIDATION_PRED_FPS_DTXSID = "massbank_validation_predfps_dtxsid.csv"


# Modeling

# Modeling input
CONFIG_FILE = "config.yaml"
CONFIG_MODELS_FILE = "config_classification.yaml"
PYTCPL_INPUT_FOLDER = "pytcpl"

REMOTE_DATA_DIR_PATH = os.path.join(ROOT_DIR, "../../../../pytcpl_eliza/data/")
REMOTE_PYTCPL_FOLDER = "output_final/"

ICE_REFERENCES = "ice_reference_file_mechtargets_aeid.csv"
PYTCPL_OVERVIEW_FILE = "assay_overview_selected.csv"

# Modeling output
MODELS_RESULTS_DIR_PATH = os.path.join(RESULTS_DIR_PATH, "models")
SELECTED_AEIDS = "selected_aeids.csv"
SELECTED_CHEMICALS = "selected_chemicals.csv"


# Prediction

# Prediction input
# < no new input files >

# Prediction output
# < no new output folder >



from pathlib import Path

import click

import mlinvitrotox.utils.predict as predict
from mlinvitrotox.utils.model import Model


@click.command("run")
@click.option(
    "--model", "-m", required=True, help="name of the model or path to model file"
)
@click.option(
    "--input_file",
    "-i",
    help="input file",
    required=True,
    type=click.Path(exists=True, path_type=Path),
)
@click.option(
    "--output_folder",
    "-o",
    required=True,
    type=click.Path(path_type=Path),
    help="path to directory where the processed output should be stored. It will be created if it does not exist.",
)
def run_models(input_file, model, output_folder):
    model_instance = Model()
    model_path = Path(model)
    model_instance.load(model_path)

    # load predicted fingerprints and get chemical ids
    df_predfps = predict.load_application_data(input_file)
    df_predfps.set_index("chem_id", inplace=True)

    # Collect predictions for each aeid

    # from pyinstrument import Profiler
    # profiler = Profiler()
    # profiler.start()

    df_predictions = model_instance.predict(df_predfps)

    # profiler.stop()
    # print(profiler.output_text(unicode=True,color=True))
    # with open('tmp/prof.html', 'w') as f:
    #     f.write(profiler.output_html())

    # set file name for predictions output file
    output_predictions_file = output_folder / f"{model_path.stem}_predictions.csv"

    # Save predictions
    df_predictions = predict.sort_by_aeid_and_chem_id(df_predictions)
    df_predictions.round(5).to_csv(output_predictions_file, index=False)
    print(f"Predictions stored to {output_predictions_file}")


if __name__ == "__main__":
    run_models()



# Data processing: 1) extract and filter massbank data

import os
import re

from pathlib import Path

import pandas as pd
import numpy as np

import time

from mlinvitrotox.constants import (
    # IDs
    CHEM_ID,
    MASSBANK_ID,
    SMILES_ID,
)


def create_and_store_unique_guids(file_path, n_files):
    """
    create and store unique (and reproducible) IDs for the MassBank data

    """

    import random
    import uuid

    rd = random.Random()
    rd.seed(0)
    list_guids = [uuid.UUID(int=rd.getrandbits(128), version=4) for _ in range(n_files)]
    df_guid = pd.DataFrame(list_guids, columns=[MASSBANK_ID])
    df_guid.to_csv(file_path, index=False)


def read_and_extract_data(file_path, selected_keys):
    """
    helper function to read MassBank files and extract relevant information


    """

    # Open file
    with open(file_path, "r") as file:
        content = file.readlines()

    # Create a dictionary to store key-value pairs with unique keys
    key_value_dict = {}
    key_count = {}

    for line in content:
        # Use regular expression to extract key-value pairs
        match = re.match(r"([^:]+):\s*([^\n]+)", line)
        if match:
            key, value = match.groups()
            # Append a unique identifier to the key
            key_identifier = key_count.get(key, 0)
            key_count[key] = key_identifier + 1
            key = f"{key}_{key_identifier:02d}"
            if key in key_value_dict:
                key_value_dict[key] += f",{value.strip()}"
            else:
                key_value_dict[key] = value.strip()

    # Filter for rows containing selected keys
    key_value_dict_selected = {
        k: v
        for k, v in key_value_dict.items()
        if any(selected_key in k for selected_key in selected_keys)
    }

    return key_value_dict_selected


def process_files(input_directory, max_files=None):
    """
    function to process all MassBank txt files and compile them in a data frame


    """

    # Set relevant keys
    selected_keys = [
        "ACCESSION",
        "RECORD_TITLE",
        "CH$NAME",
        "CH$FORMULA",
        "CH$EXACT_MASS",
        "CH$SMILES",
        "CH$IUPAC",
        "CH$LINK",
        "AC$INSTRUMENT",
        "AC$MASS_SPECTROMETRY",
        "AC$CHROMATOGRAPHY",
        "MS$FOCUSED_ION",
    ]

    # initialize
    data_list = []

    # Get all text files in the input directory
    all_files = [f for f in os.listdir(input_directory) if f.endswith(".txt")]

    # Limit the number of files to process
    if max_files is not None:
        total_files = min(len(all_files), max_files)
    else:
        total_files = len(all_files)

    # Iterate over each text file in the input directory up to the limit
    for file_number, file_name in enumerate(all_files[:total_files], start=1):
        file_path = Path(input_directory) / file_name

        # Print progress
        print(f"Processing file {file_number} of {total_files}: {file_name}")

        # Extract key-value pairs for selected keys from the file
        key_value_dict = read_and_extract_data(file_path, selected_keys)

        # Append the result to the list
        data_list.append(key_value_dict)

    # Create a DataFrame from the list of dictionaries
    df = pd.DataFrame(data_list)

    # Set base columns
    base_columns = [
        "CH$LINK",
        "MS$DATA_PROCESSING",
        "CH$NAME",
        "AC$INSTRUMENT",
        "AC$MASS_SPECTROMETRY",
        "MS$FOCUSED_ION",
        "AC$CHROMATOGRAPHY",
    ]

    # Process base columns
    for base_column in base_columns:
        # get corresponding full keys
        base_keys = [key for key in df.columns if key.startswith(base_column)]

        # join entries for each bsae column
        df[base_column] = df[base_keys].apply(
            lambda row: ",".join(row.dropna()), axis=1
        )

        # drop initial columns
        df = df.drop(columns=base_keys)

    return df


def filter_data(
    result_df,
    sirius_training_df,
    df_guid,
    output_file_structures_validation,
    output_file_extracted_massbank,
    output_file_guid_dtxsid,
):
    """
    function to filter MassBank dataframe

    """

    t0 = time.perf_counter()

    # replace some string parts
    result_df.columns = result_df.columns.str.replace("_00", "")

    # extract parts of some strings from CH$LINK
    link_extract = result_df["CH$LINK"].str.extract(
        r"CAS (?P<CAS>[^,]+)?,?"
        r".*?PUBCHEM CID:(?P<PUBCHEM>[^,]+)?,?"
        r".*?INCHIKEY (?P<INCHIKEY>[^,]+)?,?"
        r".*?CHEMSPIDER (?P<CHEMSPIDER>[^,]+)?"
        # r".*?COMPTOX (?P<COMPTOX>[^,]+)?"
    )

    # extract parts of some strings from AC$MASS_SPECTROMETRY
    # TODO try again to reduce filtering time
    # (split in two parts to reduce execution time from 8 minutes to 1 minute)
    mass_spectrometry_extract = result_df["AC$MASS_SPECTROMETRY"].str.extract(
        r"MS_TYPE (?P<MS_TYPE>[^,]+)?,?"
        r".*?ION_MODE (?P<ION_MODE>[^,]+)?,?"
        r".*?IONIZATION (?P<IONIZATION>[^,]+)?,?"
        r".*?FRAGMENTATION_MODE (?P<FRAGMENTATION_MODE>[^,]+)?,?"
        # )
        # mass_spectrometry_extract2 = result_df["AC$MASS_SPECTROMETRY"].str.extract(
        r".*?COLLISION_ENERGY (?P<COLLISION_ENERGY>[^,]+)?,?"
        r".*?RESOLUTION (?P<RESOLUTION>[^,]+)?,?"
        r".*?MASS_RANGE_M/Z (?P<MASS_RANGE_M_Z>[^,]+)?"
    )

    t1 = time.perf_counter()
    print(f"Extract and concatenate {t1 - t0:0.4f} seconds")

    # extract parts of some strings from MS$FOCUSED_ION
    focused_ion_extract = result_df["MS$FOCUSED_ION"].str.extract(
        r"BASE_PEAK (?P<BASE_PEAK>[^,]+)?,?"
        r".*?PRECURSOR_M/Z (?P<PRECURSOR_M_Z>[^,]+)?,?"
        r".*?PRECURSOR_TYPE (?P<PRECURSOR_TYPE>[^,]+)?"
    )

    # concatenate
    result_df = pd.concat(
        [result_df, link_extract, mass_spectrometry_extract, focused_ion_extract],
        axis=1,
        # [result_df, link_extract, mass_spectrometry_extract, mass_spectrometry_extract2, focused_ion_extract], axis=1
    )

    # extract DTXSID
    result_df[CHEM_ID] = result_df["CH$LINK"].str.extract(r"(DTXSID\d+)")

    # drop some columns
    result_df = result_df.drop(
        columns=[
            "MS$DATA_PROCESSING",
            "MS$FOCUSED_ION",
            "AC$MASS_SPECTROMETRY",
            "CH$LINK",
        ]
    )

    # rename some columns
    result_df.columns = [col.replace("CH$", "") for col in result_df.columns]
    result_df.columns = [col.replace("AC$", "") for col in result_df.columns]

    # unify NAs
    result_df.replace(["N/A", "na", "-", " "], np.nan, inplace=True)

    # drop SMILES columns
    result_df.dropna(subset=[SMILES_ID], inplace=True)

    # only keep certain instrumental conditions
    instrumental_strings = ["ESI", "APCI", "APPI", "MALDI"]
    instrument_regex = "|".join(instrumental_strings)
    instrument_condition = result_df["INSTRUMENT"].str.contains(
        instrument_regex, case=False, na=False
    )
    gc_condition = result_df["INSTRUMENT"].str.contains("GC", case=False, na=False)
    filtered_df = result_df[instrument_condition & ~gc_condition]
    filtered_df = filtered_df.dropna(subset=[CHEM_ID])

    # add a column indicating the presence of SIRIUS training compounds
    filtered_df = filtered_df.copy()
    filtered_df.loc[:, "sirius_training"] = filtered_df[CHEM_ID].apply(
        lambda x: "unsafe" if x in sirius_training_df[CHEM_ID].values else "safe"
    )

    # get df with unique smiles
    unique_smiles_df = (
        filtered_df.groupby(SMILES_ID)["ACCESSION"].apply(list).reset_index()
    )
    unique_smiles_df = unique_smiles_df.rename(columns={"ACCESSION": "ACCESSION_LIST"})
    unique_smiles_df["ACCESSION_LIST"] = unique_smiles_df["ACCESSION_LIST"].apply(
        lambda x: sorted(x)
    )

    # assign GUID
    unique_smiles_df[MASSBANK_ID] = df_guid[MASSBANK_ID][
        : len(unique_smiles_df)
    ].to_list()

    # store MassBank output grouped by SMILES
    unique_smiles_df = unique_smiles_df.sort_values(SMILES_ID)
    unique_smiles_df.to_csv(output_file_structures_validation, index=False)

    # get entries per ACCESSION back
    tidy_smiles_df = unique_smiles_df.explode("ACCESSION_LIST")
    tidy_smiles_df = tidy_smiles_df.rename(columns={"ACCESSION_LIST": "ACCESSION"})

    # store complete MassBank output (by ACCESSION)
    tidy_smiles_subset = tidy_smiles_df[["ACCESSION", MASSBANK_ID]]
    merged_df = pd.merge(filtered_df, tidy_smiles_subset, on="ACCESSION", how="inner")
    merged_df = merged_df.sort_values("ACCESSION")
    merged_df.to_csv(output_file_extracted_massbank, index=False)

    # count duplicate ids
    massbank_ref_file = merged_df.dropna(subset=[CHEM_ID])
    massbank_ref_file = massbank_ref_file[[CHEM_ID, MASSBANK_ID]].drop_duplicates()
    count_chem_id_per_massbank = massbank_ref_file.groupby(MASSBANK_ID)[
        CHEM_ID
    ].nunique()
    multiple_chem_id_count = count_chem_id_per_massbank[
        count_chem_id_per_massbank > 1
    ].count()
    print(f"How many GUIDs in total? {massbank_ref_file[MASSBANK_ID].nunique()} ")
    print(f"How many GUIDs have multiple DTXSID? {multiple_chem_id_count} ")

    # store GUID DTXSID matching
    massbank_ref_file = massbank_ref_file.reset_index(drop=True)
    massbank_ref_guid_dtxsid_duplicates = massbank_ref_file.duplicated().sum()
    print(massbank_ref_guid_dtxsid_duplicates)
    massbank_ref_guid_dtxsid = massbank_ref_file.drop_duplicates()
    print(massbank_ref_guid_dtxsid[CHEM_ID].nunique())
    massbank_ref_guid_dtxsid = massbank_ref_guid_dtxsid.sort_values(CHEM_ID)
    massbank_ref_guid_dtxsid.to_csv(output_file_guid_dtxsid, index=False)
    print(massbank_ref_guid_dtxsid.shape)



from pathlib import Path

import click

from mlinvitrotox.utils.model import Model

from mlinvitrotox.constants import (
    # Paths
    MODELS_RESULTS_DIR_PATH,  # 31
)


@click.command("export")
@click.option(
    "--model",
    "-m",
    required=True,
    default="2024-08-09_20-08-13",
    help="name of the model",
)
@click.option(
    "--output", "-o", required=True, help="output filename (.itox) or directory"
)
def export_model(model, output):
    model_instance = Model()
    logs_folder = Path(MODELS_RESULTS_DIR_PATH)
    model_path = Path(logs_folder / model)
    model_instance.load(model_path)
    model_instance.export(output)


if __name__ == "__main__":
    export_model()



from tqdm import tqdm

from pathlib import Path
import click

import pandas as pd

import mlinvitrotox.utils.predict as predict

from mlinvitrotox.constants import (
    # Paths
    MODELS_RESULTS_DIR_PATH,  # 31
    FPS_OUTPUT_DIR_PATH,
    # Files
    PRED_FPS_APPLICATION,  # 16
)

from mlinvitrotox.utils.model import Model


@click.command("eval")
@click.option("--model", "-m", default="2024-08-09_20-08-13", help="name of the model")
def run_evaluation(model):
    model_instance = Model()
    logs_folder = Path(MODELS_RESULTS_DIR_PATH)
    run_folder = model

    model_path = Path(logs_folder / run_folder)
    model_instance.load(model_path)

    # Get predicted application fingerprints for similarity comparison
    fps_output_directory = Path(FPS_OUTPUT_DIR_PATH)
    predfps_output_file = fps_output_directory / PRED_FPS_APPLICATION
    predfps_df = predict.load_application_data(predfps_output_file)
    predfps_df.set_index("chem_id", inplace=True)

    print("Predictions")
    predictions_df = model_instance.predict(predfps_df)
    predictions_df = predictions_df.sort_values("aeid")

    metrics_dfs_list = []

    # for each aeid
    print("Loading metric files")
    aeids = model_instance.df_assay_info["aeid"].astype("str").to_list()
    for aeid in tqdm(aeids, desc="Processing AEIDs"):
        aeid_path = (
            Path(model_instance.target_variable)
            / model_instance.ml_algorithm
            / str(aeid)
        )
        preproc_path = aeid_path / "Feature_Selection_RandomForestClassifier"

        # load metrics
        metrics_path = preproc_path / "XGBClassifier" / "test" / "metrics.csv"
        metrics_path = model_instance.get_path(metrics_path)
        met_df = pd.read_csv(metrics_path)
        met_df.insert(loc=0, column="aeid", value=aeid)
        metrics_dfs_list.append(met_df)

    # concatenate metrics and predictions
    metrics_df = pd.concat(metrics_dfs_list)
    metrics_df = metrics_df.sort_values("aeid")

    # Save predictions and metrics
    output_predictions_file = logs_folder / run_folder / "modeltraining_predictions.csv"
    output_metrics_file = logs_folder / run_folder / "modeltraining_metrics.csv"

    predictions_df = predict.sort_by_aeid_and_chem_id(predictions_df)
    predictions_df.round(5).to_csv(output_predictions_file, index=False)
    metrics_df.round(5).to_csv(output_metrics_file, index=False)
    print("Predictions and metrics stored.")


if __name__ == "__main__":
    run_evaluation()



# Data processing: 4) massbank validation of sirius

import pandas as pd
import numpy as np

import seaborn as sns
import matplotlib
from matplotlib import pyplot as plt

from sklearn.metrics import (
    accuracy_score,
    precision_score,
    recall_score,
    f1_score,
    matthews_corrcoef,
)

from mlinvitrotox.constants import (
    # IDs
    MASSBANK_ID,
)

# Setup
matplotlib.use("Agg")


def load_massbank_val_df(input_path):
    # Load dataframe
    df = pd.read_parquet(input_path)
    # df = pd.read_csv(input_path)

    return df


# Identify shared GUIDs
def filter_mbval_dfs(df_true, df_pred):
    """
    Filter MassBank validation dataframs to only contain shared chemicals (by GUID)

    """

    # get shared GUIDs between the two data frames
    true_fps_guids = set(df_true[MASSBANK_ID])
    pred_fps_guids = set(df_pred[MASSBANK_ID])
    shared_guids = true_fps_guids.intersection(pred_fps_guids)
    shared_guids_list = list(shared_guids)

    # Filter and sort dataframes (only keep entries with shared GUIDs)
    df_true_filtered = df_true[
        df_true[MASSBANK_ID].isin(shared_guids_list)
    ].sort_values(by=MASSBANK_ID)
    df_pred_filtered = df_pred[
        df_pred[MASSBANK_ID].isin(shared_guids_list)
    ].sort_values(by=MASSBANK_ID)

    # Set GUID as index
    df_true_filtered.set_index(MASSBANK_ID, inplace=True)
    df_pred_filtered.set_index(MASSBANK_ID, inplace=True)

    return df_true_filtered, df_pred_filtered


# Define tanimoto coefficient function
def tanimoto_coefficient(fp1, fp2):
    """
    helper function to calculate Tanimoto coefficient of two fingerprints

    """
    fp1 = np.array(fp1)
    fp2 = np.array(fp2)
    intersection = np.sum(np.logical_and(fp1, fp2))
    union = np.sum(np.logical_or(fp1, fp2))
    if union == 0:
        return 0.0

    return intersection / float(union)


# Calculate metrics for each GUID
def calculate_metrics_per_chemical(df_true, df_pred, output_metrics):
    # initialize
    df = pd.DataFrame(
        index=df_true.index,
        columns=["Tanimoto", "Accuracy", "Precision", "Recall", "F1", "MCC"],
    )

    # Define columns to consider (excluding GUID as it is now an index)
    fp_columns = [col for col in df_true.columns if col.isdigit()]

    # calculate metrics for each chemical
    for guid in df_true.index:
        fps_true = df_true.loc[guid, fp_columns].astype(int)
        fps_pred = df_pred.loc[guid, fp_columns].astype(int)

        y_true = fps_true.to_numpy().flatten()
        y_pred = fps_pred.to_numpy().flatten()

        df.loc[guid, "Tanimoto"] = tanimoto_coefficient(y_true, y_pred)
        df.loc[guid, "Accuracy"] = (
            accuracy_score(y_true, y_pred)
            if np.any(y_true) or np.any(y_pred)
            else np.nan
        )
        df.loc[guid, "Precision"] = precision_score(y_true, y_pred, zero_division=0)
        df.loc[guid, "Recall"] = recall_score(y_true, y_pred, zero_division=0)
        df.loc[guid, "F1"] = f1_score(y_true, y_pred, zero_division=0)
        df.loc[guid, "MCC"] = (
            matthews_corrcoef(y_true, y_pred)
            if np.any(y_true) or np.any(y_pred)
            else np.nan
        )
        df.select_dtypes(include=[np.number]).astype(float)

    # Save metrics to file
    df.rename_axis(MASSBANK_ID).sort_index().reset_index().round(5).to_csv(
        output_metrics, index=False
    )

    return df


def generate_discrepancy_figure(df_true, df_pred, output_figure):
    # Initialize an empty DataFrame for discrepancies
    discrepancy_matrix = pd.DataFrame(0, index=df_true.index, columns=df_true.columns)

    # Iterate through each GUID and column to find discrepancies
    for guid in df_true.index:
        for column in df_true.columns:
            true_bit = df_true.at[guid, column]
            pred_bit = df_pred.at[guid, column]

            if true_bit != pred_bit:
                discrepancy_matrix.at[guid, column] = -1 if true_bit == 1 else 1

    # Plot the discrepancy matrix using a heatmap
    fig, ax = plt.subplots(figsize=(15, 10))
    ax.set_title(
        "Discrepancies between true and predicted fingerprint bits", fontsize=16
    )
    ax.set_xlabel("Fingerprint index")
    ax.set_ylabel("GUID")
    ax.set_yticklabels([])

    sns.heatmap(discrepancy_matrix, cmap=["blue", "white", "green"], cbar=False, ax=ax)

    plt.xticks(rotation=90)
    plt.tight_layout()
    plt.savefig(output_figure, format="png")
    plt.close("all")


def calculate_metrics_per_bit(df_true, df_pred, output_metrics):
    # initialize
    df = pd.DataFrame(
        columns=["Accuracy", "Precision", "Recall", "F1", "MCC"],
        index=df_true.columns,
    )

    # calculate metrics for each column
    for column in df_true.columns:
        y_true = df_true[column].to_numpy()
        y_pred = df_pred[column].to_numpy()

        # Calculate metrics for each column
        accuracy = (
            accuracy_score(y_true, y_pred)
            if np.any(y_true) or np.any(y_pred)
            else np.nan
        )
        precision = precision_score(y_true, y_pred, zero_division=0)
        recall = recall_score(y_true, y_pred, zero_division=0)
        f1 = f1_score(y_true, y_pred, zero_division=0)
        mcc = (
            matthews_corrcoef(y_true, y_pred)
            if np.any(y_true) or np.any(y_pred)
            else np.nan
        )

        # Assign metrics to the DataFrame
        df.at[column, "Accuracy"] = accuracy
        df.at[column, "Precision"] = precision
        df.at[column, "Recall"] = recall
        df.at[column, "F1"] = f1
        df.at[column, "MCC"] = mcc

    # store output
    df.rename_axis("absoluteIndex").sort_index().reset_index().round(5).to_csv(
        output_metrics, index=False
    )

    return df


def plot_metrics_histogram(df_column_metrics, output_figure):
    plt.figure(figsize=(10, 6))
    colors = ["red", "green", "blue", "purple", "orange"]
    alpha = 0.5
    bins = 20

    for i, metric in enumerate(df_column_metrics.columns):
        plt.hist(
            df_column_metrics[metric].dropna(),
            bins=bins,
            alpha=alpha,
            label=metric,
            color=colors[i],
        )
    plt.title("Histogram of metrics across fingerprint bits")
    plt.xlabel("Metric value")
    plt.ylabel("Frequency")
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig(output_figure, format="png")
    plt.close()


def plot_metrics_density(df_column_metrics, output_figure):
    plt.figure(figsize=(10, 6))
    colors = ["red", "green", "blue", "purple", "orange"]

    for i, metric in enumerate(df_column_metrics.columns):
        sns.kdeplot(
            df_column_metrics[metric].dropna(), color=colors[i], label=metric, fill=True
        )
    plt.title("Density distribution of metrics across fingerprint bits")
    plt.xlabel("Metric value")
    plt.ylabel("Density")
    plt.legend()
    plt.tight_layout()
    plt.savefig(output_figure, format="png")
    plt.close()


def save_indexes_passed_validation(
    df_metrics_bit, threshold_recall, threshold_precision, output_file
):
    passed_validation = df_metrics_bit[
        (df_metrics_bit["Recall"] > threshold_recall)
        & (df_metrics_bit["Precision"] > threshold_precision)
    ]
    indexes_passed = pd.DataFrame(passed_validation.index, columns=["absoluteIndex"])
    print(f"Original # indexes {df_metrics_bit.shape}")
    print(f"Filtered # indexes {indexes_passed.shape}")

    # store output
    indexes_passed.to_csv(output_file, index=False)



dict_rename = {
    "hitcall": "hitcall",
    "hitcall_c": "hitcall (c)",
    "True": "positive",
    "False": "negative",
    "macro avg": "macro avg",
    "weighted avg": "weighted avg",
    "test": "Internal validation",
    "mb_val_structure": "MB validation from structure",
    "mb_val_sirius": "MB validation SIRIUS-predicted",
    "Internal validation": "Internal",
    "MB validation from structure": "MB structure",
    "MB validation SIRIUS-predicted": "MB SIRIUS",
    "default": "default=0.5",
    "tpr": "TPR≈0.5",
    "tnr": "TNR≈0.5",
    "optimal": "cost(TPR,TNR)",
    "XGBClassifier": "XGBoost",
    "XGBoost": "XGB",
    "RF": "RF",
    "RBF SVM": "SVM",
    "MLP": "MLP",
    "LR": "LR",
    "RandomForestClassifier": "RF",
    "LogisticRegression": "LR",
    "SVC": "RBF SVM",
    "MLPClassifier": "MLP",
    "accuracy": "accuracy",
}

acronym_mapping = {
    "Cell Cycle": "CP",
    "Not Annotated": "na",
    "not_annotated": "na",
    "Constitutive Androstane Receptor Modulation": "AN",
    "PPARG": "PPARG",  # Peroxisome proliferator activated receptor gamma
    "Pregnane X Receptor Modulation": "PXR",
    "RAR-related Orphan Receptor Modulation": "ROR",
    "Other Transcription Factors": "TF",
    "Other Developmental Signaling Transcription Factors": "TF",
    "Oxidative Stress": "OSR",
    "Inflammation": "INF",
    "p53 Modulation": "GTX",
    "Farnesoid X-activated Receptor Modulation": "FXR",
    "Peroxisome Proliferator Activated Receptor Modulation": "PPARG",
    "Vitamin D Modulation": "IM",
    "DNA Repair": "GTX",
    "Estrogen Receptor Modulation": "ER",
    "Liver X Receptor Modulation": "LXR",
    "TP53": "GTX",
    "Retinoic Acid Receptor Modulation": "RAR",
    "Retinoid X Receptor Modulation": "RXR",
    "Thyroid Receptor Modulation": "THR",
    "ICAM1": "IM",
    "VascularTissue": "CP",
    "Cell Growth": "CP",
    "TF": "TF",
    "Extracellular Matrix Degradation": "CP",
    "Vascularization": "CP",
    "Androgen Receptor Modulation": "AR",
    "VEGF": "PR",
    "Histone Modification": "GTX",
    "Mt Dys": "MF",
    "t-PA": "PR",
    "Estrogen-related Receptor Modulation": "ER",
    "IL6": "INF",
    "Progesterone Receptor Modulation": "P4",
    "SAA1": "PR",
    "Clotting": "CP",
    "Gene Expression Regulation, Xenobiotic Response": "XNR",
    "Apoptosis": "APO",
    "PAI -1(SERPINE1)": "PR",
    "Proliferation": "CP",
    "Cell Morphology": "CP",
    "Mitochondrial Function": "MF",
    "Aryl Hydrocarbon Receptor Modulation": "AhR",
    "ER Alpha": "ER",  # Same acronym as "Estrogen Receptor Modulation"
    "Aromatase Activity Modulation": "AA",
    "Glucocorticoid Receptor Modulation": "GC",
    "Malformation": "CP",
    "Receptors, Thyrotropin-Releasing Hormone": "TRH",
    "Neurotransmission": "NR",
    "DNA Damage": "GTX",
    "NR3C1": "GC",
    "Progesterone-related Steroid Hormone Metabolism": "P4",
    "Progestogen Biosynthesis and Metabolism": "NRC",
    "Androgen-related Steroid Hormone Metabolism": "AR",
    "Glucocorticoid Biosynthesis and Metabolism": "GC",
    "Glucocorticoid-related Steroid Hormone Metabolism": "GC",
    "Estrogen Biosynthesis and Metabolism": "ER",
    "Androgen Biosynthesis and Metabolism": "AR",
    "Sodium/Iodide Cotransporter": "NIS",
    "Pregnane X Receptor Modulation": "PXR",
    "Constitutive Androstane Receptor Modulation": "AN",
}


def _blank_to_underscore(x):
    return x.replace(" ", "_")


def predict_for_endpoint(endpoint, clf_data, features):
    feature_selection_model = clf_data[0]
    features = feature_selection_model.transform(features)
    clf = clf_data[1]
    prediction = clf.predict(features)
    return endpoint, prediction


def pandas_df_to_latex(data, caption="", label=""):
    # Initialize the LaTeX table with the longtable environment
    latex_table = f'\\begin{{longtable}}{{{"l" * len(data.columns)}}}\n'
    latex_table += (
        "\\caption{"
        + caption
        + "}\\label{tab:"
        + label.lower().replace(" ", "_")
        + "}\\\\\n"
    )
    latex_table += "\\toprule\n\\midrule\n"

    # Extract the header and remove it from the DataFrame
    header = data.columns
    data = data.values

    # Add the header row with \small to the LaTeX table
    latex_table += " & ".join(["\\small " + col for col in header]) + "\\\\\n\\hline\n"

    # Iterate through the data rows and add them to the table
    for row in data:
        latex_table += " & ".join(map(str, row)) + "\\\\\n"

    # Add table footer and replace \endtabular with \endlongtable
    latex_table += "\\bottomrule\n\\end{longtable}"
    latex_table = latex_table.replace("\\endtabular", "\\endlongtable")

    return latex_table



import os
import sys
import re
import shutil

from pathlib import Path

import logging
from datetime import datetime
import joblib
import yaml
from tqdm import tqdm

import numpy as np
import pandas as pd

import matplotlib
from matplotlib import pyplot as plt

from sklearn.pipeline import Pipeline
from sklearn.base import BaseEstimator, TransformerMixin
from xgboost import XGBClassifier  # needed for build_pipeline()
# from sklearn.linear_model import LogisticRegression
# from sklearn.svm import SVC
# from sklearn.neural_network import MLPClassifier
# from sklearn.ensemble import RandomForestClassifier

from mlinvitrotox.constants import (
    # IDs
    CHEM_ID,
)

# Plotting backend
matplotlib.use("Agg")

LOGGER = logging.getLogger(__name__)


def empty_pytcpl_input_directory(
    pytcpl_input_folder,
):
    for filename in os.listdir(pytcpl_input_folder):
        file_path = Path(pytcpl_input_folder) / filename
        if os.path.isfile(file_path):
            os.unlink(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path)


def copy_pytcpl_parquet_files(
    remote_pytcpl_folder,
    pytcpl_input_folder,
):
    for filename in os.listdir(remote_pytcpl_folder):
        if filename.endswith(".parquet.gzip"):
            parquet_path = Path(remote_pytcpl_folder) / filename
            shutil.copy2(parquet_path, pytcpl_input_folder)
            print(f"Copied {parquet_path} to {pytcpl_input_folder}")


def load_config(config_path, logs_folder):
    global START_TIME, LOG_PATH

    config = load_simple_config(config_path)

    START_TIME = datetime.now()
    LOGGER, RUN_FOLDER = init_logger(logs_folder)
    log_config_path = Path(LOG_PATH) / "log/config.yaml"
    with open(log_config_path, "w") as file:
        yaml.dump(config, file)
        LOGGER.info(f"Config file dumped to '{log_config_path}'")

    return config, START_TIME, LOGGER, RUN_FOLDER, LOG_PATH


def load_simple_config(config_path):
    with open(config_path, "r") as file:
        config = yaml.safe_load(file)
        if config["ignore_warnings"]:
            import warnings

            warnings.filterwarnings("ignore")

    return config


def init_target_variable(target_variable):
    global TARGET_VARIABLE, DUMP_FOLDER
    TARGET_VARIABLE = target_variable
    DUMP_FOLDER = os.path.join(str(LOG_PATH), str(TARGET_VARIABLE))
    os.makedirs(DUMP_FOLDER, exist_ok=True)
    return TARGET_VARIABLE


def init_ml_algo(ml_algorithm, config_models_path):
    global ML_ALGORITHM, DUMP_FOLDER
    ML_ALGORITHM = ml_algorithm

    # load config models file
    with open(config_models_path, "r") as file:
        config_estimators = yaml.safe_load(file)

    # store config models file in log folder
    config_models_file = config_models_path.split("/")[-1]
    with open(os.path.join(LOGGER_FOLDER, config_models_file), "w") as file:
        yaml.dump(config_estimators, file)

    # create dump folder
    DUMP_FOLDER = os.path.join(LOG_PATH, TARGET_VARIABLE, str(ml_algorithm))
    os.makedirs(DUMP_FOLDER, exist_ok=True)

    return config_estimators


def init_aeid(aeid):
    global AEID, DUMP_FOLDER
    AEID = str(aeid)
    DUMP_FOLDER = os.path.join(
        str(LOG_PATH), str(TARGET_VARIABLE), str(ML_ALGORITHM), str(AEID)
    )
    os.makedirs(DUMP_FOLDER, exist_ok=True)


def init_preprocessing_pipeline(preprocessing_pipeline):
    global PREPROCESSING_PIPELINE, DUMP_FOLDER
    preprocessing_pipeline_name = preprocessing_pipeline[
        -1
    ].estimator.__class__.__name__
    PREPROCESSING_PIPELINE = f"Feature_Selection_{preprocessing_pipeline_name}"
    LOGGER.info(f"Apply {PREPROCESSING_PIPELINE}..")
    DUMP_FOLDER = os.path.join(
        str(LOG_PATH),
        str(TARGET_VARIABLE),
        str(ML_ALGORITHM),
        str(AEID),
        str(PREPROCESSING_PIPELINE),
    )
    os.makedirs(DUMP_FOLDER, exist_ok=True)
    return PREPROCESSING_PIPELINE


def init_estimator_pipeline(estimator_name):
    global ESTIMATOR_PIPELINE, DUMP_FOLDER
    ESTIMATOR_PIPELINE = estimator_name
    DUMP_FOLDER = os.path.join(
        str(LOG_PATH),
        str(TARGET_VARIABLE),
        str(ML_ALGORITHM),
        str(AEID),
        str(PREPROCESSING_PIPELINE),
        str(ESTIMATOR_PIPELINE),
    )
    os.makedirs(DUMP_FOLDER, exist_ok=True)
    return ESTIMATOR_PIPELINE


def init_validation_set(validation_set):
    global VALIDATION_SET, DUMP_FOLDER
    VALIDATION_SET = validation_set
    LOGGER.info(f"Validation Set: {VALIDATION_SET}")
    DUMP_FOLDER = os.path.join(
        str(LOG_PATH),
        str(TARGET_VARIABLE),
        str(ML_ALGORITHM),
        str(AEID),
        str(PREPROCESSING_PIPELINE),
        str(ESTIMATOR_PIPELINE),
        str(VALIDATION_SET),
    )
    os.makedirs(DUMP_FOLDER, exist_ok=True)
    return VALIDATION_SET


def merge_assay_info(pytcpl_df, ice_df):
    """
    Adapt the assay_info DataFrame:
    - Convert aeid to string
    - Keep only specified columns
    - Left join with ice_df on aeid to exclude "Cell Viability" from "MechanisticTarget" and include unique "MechanisticTarget"
    """
    col_mt = "MechanisticTarget"

    # prepare
    pytcpl_df["aeid"] = pytcpl_df["aeid"].astype(str)
    ice_df["aeid"] = ice_df["aeid"].astype(str)

    # remove cell viability entries
    ice_df = ice_df[ice_df[col_mt] != "Cell Viability"].copy()
    # remove weird characters
    pattern = re.compile(r"\s+")
    ice_df[col_mt] = ice_df[col_mt].apply(lambda x: re.sub(pattern, " ", str(x)))
    # remove CardioTox
    ice_df["MechanisticTarget"] = ice_df["MechanisticTarget"].str.replace(
        "CardioTox_", ""
    )
    # unify not annotated
    ice_df[col_mt] = ice_df[col_mt].replace(
        ["Not Annotated", "Unspecified", "Unassigned", "nan"], "not_annotated"
    )
    # remove duplicates
    ice_df_unique = ice_df.drop_duplicates(subset=["aeid", col_mt])

    # merge
    merged_df = pd.merge(
        pytcpl_df, ice_df_unique[["aeid", col_mt]], on="aeid", how="left"
    )
    columns_to_keep = [
        "aeid",
        "assay_component_endpoint_name",
        col_mt,
        "signal_direction",
    ]
    merged_df = merged_df[columns_to_keep]
    merged_df[col_mt] = merged_df[col_mt].fillna("not_annotated")

    return merged_df


def get_aeid_files(
    pytcpl_input_folder,
    target_variable,
    assay_info_df,
):
    col_mt = "MechanisticTarget"

    # get list of pytcpl files
    parquet_files = [
        f for f in os.listdir(pytcpl_input_folder) if f.endswith(".parquet.gzip")
    ]

    # initialize
    aeid_list = []

    for filename in tqdm(parquet_files, desc="Processing aeid files"):
        file_path = Path(pytcpl_input_folder) / filename
        df = pd.read_parquet(file_path)
        if target_variable not in df.columns:
            print(f"Target variable not in {filename}")
            continue

        aeid = filename.split(".")[0]
        aeid_list.append(aeid)

    # concatenate
    aeid_df = pd.DataFrame(aeid_list, columns=["aeid"])

    # merged with aeid info to add mechanistic target
    merged_df = pd.merge(
        aeid_df,
        assay_info_df,
        on="aeid",
        how="left",
    )
    merged_df[col_mt] = merged_df[col_mt].fillna("not_annotated")
    merged_df["aeid"] = merged_df["aeid"].astype("int")

    return merged_df


def select_aeids_for_training(
    pytcpl_input_folder,
    assay_info_df,
    aeids=None,
    mech_target=None,
):
    # all aeids from pytcpl
    available_aeids = [
        f.split(".")[0]
        for f in os.listdir(pytcpl_input_folder)
        if f.endswith(".parquet.gzip")
    ]

    # select all
    if aeids is None and mech_target is None:
        selected_aeids = available_aeids

    # select from list
    if aeids is not None:
        provided_aeids = [
            str(aeid) for aeid in (aeids if isinstance(aeids, list) else [aeids])
        ]
        selected_aeids = [aeid for aeid in provided_aeids if aeid in available_aeids]
    else:
        selected_aeids = []

    # select from mechanistic target
    if mech_target is not None:
        target_pattern = mech_target.lower()
        matched_targets = assay_info_df["MechanisticTarget"].str.contains(
            target_pattern, case=False, na=False
        )
        matched_aeids = assay_info_df[matched_targets]["aeid"].astype(str).tolist()

        if len(selected_aeids) > 0:
            selected_aeids = [aeid for aeid in matched_aeids if aeid in selected_aeids]

        else:
            selected_aeids = [aeid for aeid in matched_aeids if aeid in available_aeids]

    selected_aeids = sorted(set(selected_aeids))
    return selected_aeids


def get_fingerprint_df(fps_input_file):
    if isinstance(fps_input_file, Path) and fps_input_file.suffix == ".csv":
        fps_df = pd.read_csv(fps_input_file)
    else:
        fps_df = pd.read_parquet(fps_input_file)

    integer_columns = [col for col in fps_df.columns if col.isdigit()]
    fps_df[integer_columns] = fps_df[integer_columns].astype(int)

    columns_to_keep = [col for col in fps_df.columns if col.isdigit() or col == CHEM_ID]
    fps_df = fps_df[columns_to_keep]

    print(
        f"Dataframe with true fingerprints: {fps_df.shape[0]} chemicals, {fps_df.iloc[:, 1:].shape[1]} fingerprint bits"
    )

    return fps_df


def process_single_aeid(
    pytcpl_input_folder,
    aeid,
    target_variable,
    ice_omit_filtering,
    activity_threshold,
):
    # load file
    file_path = Path(pytcpl_input_folder) / f"{aeid}.parquet.gzip"
    if not os.path.exists(file_path):
        print(f"File not found: {file_path}")
        return None
    df = pd.read_parquet(file_path)

    # apply ice omit filtering
    # (entries with NA are to be kept)
    if ice_omit_filtering:
        keep_compound_mask = df["omit_flag"].isin(["None", np.nan])
        df_kept = df[keep_compound_mask]
        num_rows_kept = len(df_kept)
        num_rows_removed = len(df) - num_rows_kept
        print(f"ICE filtering: Number of compounds kept: {num_rows_kept}")
        print(f"ICE filtering: Number of compounds removed: {num_rows_removed}")
    else:
        df_kept = df.copy()

    # get first DSSTOX column
    dsstox_cols = df_kept.columns[df_kept.columns.str.contains("dsstox")]
    if not dsstox_cols.empty:
        dsstox_col = dsstox_cols[0]
    else:
        print("No 'dsstox' column found")
        return None

    # check for target_variable
    if target_variable not in df_kept.columns:
        print(f"'{target_variable}' column not found")
        return None

    # only keep dsstox and target variable columns
    df_selected = df_kept.loc[:, [dsstox_col, target_variable]]
    df_selected.dropna(inplace=True)

    # apply activity threshold
    df_selected.loc[:, target_variable] = (
        df_selected[target_variable] >= activity_threshold
    ).astype(int)

    # merge entries for duplicate dsstox entries
    # TODO do we have duplicate dsstox entries?
    n_before = df_selected.shape[0]
    print(f"Before grouping by dsstox_col: Number of rows: {n_before}")
    df_selected = df_selected.groupby(dsstox_col).max().reset_index()
    n_after = df_selected.shape[0]
    print(f"After grouping by dsstox_col: Number of rows: {n_after}")
    if n_after != n_before:
        print("Chemicals merged through groupby by dsstox_col")

    # rename DSSTOX column
    df_selected = df_selected.rename(columns={dsstox_col: CHEM_ID})
    return df_selected


def merge_assay_and_fingerprint_df(assay_df, fps_df, selected_chemicals_file=None):
    assert CHEM_ID in assay_df.columns, f"No '{CHEM_ID}' column found in assay_df"
    assert CHEM_ID in fps_df.columns, f"No '{CHEM_ID}' column found in fps_df"
    merged_df = pd.merge(
        assay_df,
        fps_df,
        left_on=CHEM_ID,
        right_on=CHEM_ID,
        how="inner",
    ).reset_index(drop=True)
    assert (
        merged_df.shape[0] == merged_df[CHEM_ID].nunique()
    ), "Merging resulted in duplicate rows"

    if selected_chemicals_file is not None:
        merged_df[CHEM_ID].to_csv(
            Path(DUMP_FOLDER) / selected_chemicals_file, index=False
        )
    return merged_df


def cluster_threshold_all(
    df,
    n_clusters,
    random_state,
):
    from sklearn.cluster import KMeans

    # np.random.seed(random_state)
    if df.index.name != CHEM_ID:
        df = df.set_index(CHEM_ID)

    if "Cluster" in df.columns:
        clustering_features = df.drop(columns=["Cluster"]).copy()
    else:
        clustering_features = df.copy()
    kmeans = KMeans(n_clusters=n_clusters, random_state=random_state).fit(
        clustering_features
    )
    df["Cluster"] = kmeans.labels_
    return df.reset_index()


def clustering(
    df,
    target_variable,
    n_cls_0,
    n_cls_1,
    random_state,
):
    clusters_0 = []
    clusters_1 = []

    df_0 = df[df[target_variable] == 0]
    df_1 = df[df[target_variable] == 1]

    n_clusters_0 = len(df_0) // n_cls_0 + (1 if len(df_0) % n_cls_0 != 0 else 0)
    n_clusters_1 = len(df_1) // n_cls_1 + (1 if len(df_1) % n_cls_1 != 0 else 0)

    features_0 = df_0.drop(target_variable, axis=1)
    features_1 = df_1.drop(target_variable, axis=1)

    if not features_0.empty:
        df_0 = cluster_threshold_all(features_0, n_clusters_0, random_state)
        df_0[target_variable] = 0
        clusters_0.append(df_0)

    if not features_1.empty:
        df_1 = cluster_threshold_all(features_1, n_clusters_1, random_state)
        df_1[target_variable] = 1
        clusters_1.append(df_1)

    combined_df = pd.concat(clusters_0 + clusters_1, ignore_index=True)
    combined_df["Cluster_str"] = combined_df["Cluster"].astype(str)
    combined_df["target_str"] = combined_df[target_variable].astype(str)
    combined_df["clustering_rule"] = (
        combined_df["Cluster_str"] + "-" + combined_df["target_str"]
    )
    combined_df.drop(columns=["Cluster", "target_str"], inplace=True)
    return combined_df


def filter_with_sirius_quality_indexes(
    df,
    indexes_path,
    target_variable,
):
    # load selected indexes
    indexes_df = pd.read_csv(indexes_path)
    col_name = indexes_df.columns[0]
    indexes_df = indexes_df.sort_values(col_name).drop_duplicates()

    # convert to string with 4 characters
    if indexes_df["absoluteIndex"].dtype == np.int64:
        filter_indexes = indexes_df["absoluteIndex"].tolist()
    else:
        filter_indexes = indexes_df["absoluteIndex"].astype(str).tolist()
    filter_columns = [str(index).zfill(4) for index in filter_indexes]

    # add other entries and filter
    filter_columns = filter_columns + [CHEM_ID, target_variable, "clustering_rule"]
    filtered_df = df[filter_columns].copy()

    return filtered_df


def split_training_validation_data(
    filtered_df,
    pred_fps_df,
    target_variable,
):
    # get DTXSID in both data frames --> validation
    filtered_df_dtxsid = filtered_df[[CHEM_ID]]
    pred_fps_df_dtxsid = pred_fps_df[[CHEM_ID]]
    dtxsid_val = set(filtered_df_dtxsid[CHEM_ID]).intersection(
        set(pred_fps_df_dtxsid[CHEM_ID])
    )
    # get DTXSIDs not shared in data frames --> training and test
    dtxsis_train_test = set(filtered_df_dtxsid[CHEM_ID]).difference(
        set(pred_fps_df_dtxsid[CHEM_ID])
    )
    df_validation_true = filtered_df[filtered_df[CHEM_ID].isin(dtxsid_val)]
    df_validation_pred = pred_fps_df[pred_fps_df[CHEM_ID].isin(dtxsid_val)]
    df_train_test = filtered_df[filtered_df[CHEM_ID].isin(dtxsis_train_test)]
    print(f"df_validation_true {df_validation_true.shape}")
    print(f"df_validation_pred {df_validation_pred.shape}")
    print(f"df_train_test {df_train_test.shape}")
    print(f"df_train_test columns {df_train_test.columns}")

    # only keep common columns for predicted validation data frame
    common_columns = df_validation_pred.columns.intersection(df_validation_true.columns)
    df_validation_pred = df_validation_pred.loc[:, common_columns]

    # perform checks
    columns_training_test = set(
        col
        for col in df_train_test.columns
        if col != target_variable and "cluster" not in col.lower()
    )
    columns_validation_pred = set(
        col
        for col in df_validation_pred.columns
        if col != target_variable and "cluster" not in col.lower()
    )
    columns_validation_true = set(
        col
        for col in df_validation_true.columns
        if col != target_variable and "cluster" not in col.lower()
    )

    different_columns = {
        "Training/Test vs. Validation Pred": columns_training_test.symmetric_difference(
            columns_validation_pred
        ),
        "Training/Test vs. Validation True": columns_training_test.symmetric_difference(
            columns_validation_true
        ),
        "Validation Pred vs. Validation True": columns_validation_pred.symmetric_difference(
            columns_validation_true
        ),
    }

    if all(len(diff) == 0 for diff in different_columns.values()):
        print("All DataFrames have exactly the same columns (except target_variable).")
    else:
        print("DataFrames do not have the same columns (except target_variable).")
        for key, diff in different_columns.items():
            if len(diff) > 0:
                print(f"Differences in columns between {key}: {', '.join(diff)}")

    duplicates_true_exist = df_validation_true[CHEM_ID].duplicated().any()
    print(f"DTXSID duplicates in df_validation_true: {duplicates_true_exist}")
    duplicates_pred_exist = df_validation_pred[CHEM_ID].duplicated().any()
    print(f"DTXSID duplicates in df_validation_pred: {duplicates_pred_exist}")
    return df_train_test, df_validation_true, df_validation_pred


def partition_training_data(
    df_train_test,
    target_variable,
):
    # get feature_names
    feature_names = [
        col for col in df_train_test.columns if col.isdigit() and col != CHEM_ID
    ]

    # set DTXSID as index
    if CHEM_ID in df_train_test.columns:
        df_train_test = df_train_test.set_index(CHEM_ID).sort_index()

    # get X, y, and clustering rule for stratification
    X = df_train_test[feature_names]
    y = df_train_test[target_variable]  # .astype(np.uint8)
    cluster = df_train_test["clustering_rule"]

    return X, y, cluster, feature_names


def partition_validation_data(
    df_validation_true,
    df_validation_pred,
    target_variable,
    feature_names,
):
    # set DTXSID as index
    if CHEM_ID in df_validation_true.columns:
        df_validation_true = df_validation_true.set_index(CHEM_ID).sort_index()
    if CHEM_ID in df_validation_pred.columns:
        df_validation_pred = df_validation_pred.set_index(CHEM_ID).sort_index()

    # get X and y
    X_mbval_structure = df_validation_true[feature_names]
    X_mbval_sirius = df_validation_pred[feature_names]
    y_mbval = df_validation_true[target_variable].astype(np.uint8)

    return (
        X_mbval_structure,
        X_mbval_sirius,
        y_mbval,
    )


def split_training_data(
    X,
    y,
    cluster,
    train_test_split_ratio,
    random_state,
):
    from sklearn.model_selection import train_test_split

    # get single member clusters
    cluster_counts = cluster.value_counts()
    single_member_clusters = cluster_counts[cluster_counts == 1].index
    single_member_indices = [
        cluster[cluster == i].index[0] for i in single_member_clusters
    ]
    X_single = X[X.index.isin(single_member_indices)]
    y_single = y[y.index.isin(single_member_indices)]
    X_multiple = X[~X.index.isin(single_member_indices)]
    y_multiple = y[~y.index.isin(single_member_indices)]
    cluster_multiple = cluster[~cluster.index.isin(single_member_indices)]
    print(X.shape, X_single.shape, y.shape, y_single.shape)
    print(X.shape, X_multiple.shape, y.shape, y_multiple.shape)

    # stratify multiple clusters data
    stratify = cluster_multiple
    X_train, X_test, y_train, y_test = train_test_split(
        X_multiple,
        y_multiple,
        test_size=train_test_split_ratio,
        random_state=random_state,
        shuffle=True,  # shuffle the data before splitting (default)
        stratify=stratify,
    )

    # add single clusters to training data
    X_train = pd.concat([X_train, X_single], ignore_index=True)
    y_train = pd.concat([y_train, y_single], ignore_index=True)
    print(X_train.shape, y_train.shape, X_test.shape, y_test.shape)

    return X_train, y_train, X_test, y_test


def handle_oversampling(
    X,
    y,
    random_state,
):
    from imblearn.over_sampling import RandomOverSampler

    ros = RandomOverSampler(random_state=random_state)
    X, y = ros.fit_resample(X, y)
    return X, y


class RemoveCorrelatedFeatures(BaseEstimator, TransformerMixin):
    def __init__(self, threshold=0.9):
        self.threshold = threshold
        self.correlated_features = set()

    def fit(self, X, y=None):
        corr_matrix = np.corrcoef(X, rowvar=False)
        for i in range(len(corr_matrix)):
            for j in range(i + 1, len(corr_matrix)):
                if abs(corr_matrix[i, j]) > self.threshold:
                    self.correlated_features.add(j)
        return self

    def transform(self, X, y=None):
        return np.delete(X, list(self.correlated_features), axis=1)


def build_preprocessing_pipeline(
    feature_selection,
    remove_lowvariance_features,
    remove_highcorrelation_features,
    variance_threshold,
    correlation_threshold,
    random_state,
):
    from sklearn.feature_selection import VarianceThreshold
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.feature_selection import SelectFromModel

    preprocessing_pipeline_steps = []
    if feature_selection:
        if remove_lowvariance_features:
            # remove all low-variance features
            feature_selection_variance_threshold = VarianceThreshold(variance_threshold)
            preprocessing_pipeline_steps.append(
                (
                    "feature_selection_variance_threshold",
                    feature_selection_variance_threshold,
                )
            )

        if remove_highcorrelation_features:
            # remove correlated features
            remove_correlated_features = RemoveCorrelatedFeatures(
                threshold=correlation_threshold
            )
            preprocessing_pipeline_steps.append(
                (
                    "remove_correlated_features",
                    remove_correlated_features,
                )
            )

        feature_selection_model = RandomForestClassifier(random_state=random_state)

        feature_selection_from_model = SelectFromModel(
            estimator=feature_selection_model, threshold="mean"
        )

        preprocessing_pipeline_steps.append(
            ("feature_selection_from_model", feature_selection_from_model)
        )

    preprocessing_pipeline = Pipeline(preprocessing_pipeline_steps)
    LOGGER.info("Built Preprocessing pipeline (feature selection)")
    return [preprocessing_pipeline]


def preprocess_all_sets(
    preprocessing_pipeline,
    feature_names,
    X_train_os,
    y_train_os,
    X_train,
    X_test,
    X_mbval_structure,
    X_mbval_sirius,
):
    # Feature selection fitted on train set. Transform all sets with the same feature selection
    if preprocessing_pipeline.steps:
        X_train_os = preprocessing_pipeline.fit_transform(
            X_train_os,
            y_train_os,
        )

        num_features = X_train_os.shape[1]
        LOGGER.info(f"Number of selected features: {num_features}")

        # Get the selected feature indices and then names
        selected_feature_indices = preprocessing_pipeline[-1].get_support()
        selected_feature_names = [
            feature_names[i]
            for i, selected in enumerate(selected_feature_indices)
            if selected
        ]
        selected_feature_df = pd.DataFrame(selected_feature_names, columns=["feature"])
        selected_feature_df.to_csv(
            Path(DUMP_FOLDER) / "selected_features.csv", index=False
        )

        # Transform other sets (e.g. subselect feature columns that were selected by the feature selection model)
        X_train = preprocessing_pipeline.transform(X_train)
        X_test = preprocessing_pipeline.transform(X_test)
        X_mbval_structure = preprocessing_pipeline.transform(X_mbval_structure)
        X_mbval_sirius = preprocessing_pipeline.transform(X_mbval_sirius)

        # Assert that all sets have the same features
        if (
            num_features != X_train.shape[1]
            or num_features != X_test.shape[1]
            or num_features != X_mbval_structure.shape[1]
            or num_features != X_mbval_sirius.shape[1]
        ):
            LOGGER.error(
                f"Number of features in train, test, mbval_structure, mbval_sirius do not match: "
                f"{num_features}, {X_train.shape[1]}, {X_test.shape[1]}, {X_mbval_structure.shape[1]}, {X_mbval_sirius.shape[1]}"
            )
            raise RuntimeError("Error in feature selection")

    # Save preprocessing_model
    preprocessing_model_path = Path(DUMP_FOLDER) / "preprocessing_model.joblib"
    joblib.dump(preprocessing_pipeline, preprocessing_model_path, compress=3)

    return (
        selected_feature_names,
        X_train_os,
        X_train,
        X_test,
        X_mbval_structure,
        X_mbval_sirius,
    )


def log_binarized_label_count(y, title, activity_threshold):
    # Ensure y is compared activity_threshold to produce a boolean Series
    y_binarized = y >= activity_threshold

    # Use value_counts() to count both True and False explicitly, ensuring no missing values
    counts = (
        y_binarized.value_counts(dropna=False)
        .reindex([False, True], fill_value=0)
        .values
    )

    # Log the counts. Note: `counts[0]` represents False (inactive), `counts[1]` represents True (active)
    LOGGER.info(
        f"Binarized Label Count {title}: {len(y)} datapoints"
        f" with {counts[0]} inactive, {counts[1]} active "
        f"({counts[1] / sum(counts) * 100:.2f}%)"
    )


def build_pipeline(estimator, random_state):
    pipeline_steps = []
    for i, step in enumerate(estimator["steps"]):
        step_name = step["name"]
        step_args = step.get("args", {})  # get the hyperparameters for the step, if any
        step_args.update({"random_state": random_state})
        step_instance = globals()[step_name](
            **step_args
        )  # dynmically create an instance of the step
        pipeline_steps.append((step_name, step_instance))

    pipeline = Pipeline(pipeline_steps)
    LOGGER.info("=" * 100 + "\n")
    LOGGER.info(f"Built Pipeline for {ESTIMATOR_PIPELINE}")
    return pipeline


def build_param_grid(estimator_steps):
    param_grid = {}
    for step in estimator_steps:
        step_name = step["name"]
        step_args = step.get("args", {})
        param_grid.update(
            {
                f"{step_name}__{key}": value
                for key, value in step_args.items()
                if isinstance(value, list)
            }
        )
    return param_grid


def perform_grid_search_cv(
    X_train, y_train, estimator, pipeline, n_splits, n_repeats, random_state
):
    from sklearn.model_selection import GridSearchCV, RepeatedStratifiedKFold

    cv = RepeatedStratifiedKFold(
        n_splits=n_splits,
        n_repeats=n_repeats,
        random_state=random_state,
    )

    grid_search_cv = GridSearchCV(
        pipeline,
        param_grid=build_param_grid(estimator["steps"]),
        cv=cv,
        scoring=None,
        n_jobs=-1,
        verbose=3,
    )

    # Use train set as input to Grid Search Cross Validation
    # (k-fold cross-validation sets drawn internally from train set)
    grid_search_cv = grid_search_cv.fit(X_train, y_train)

    LOGGER.info(f"{estimator['name']}: GridSearchCV Results:")
    best_params = (
        grid_search_cv.best_params_ if grid_search_cv.best_params_ else "default"
    )
    LOGGER.info(f"Best params: {best_params}")
    LOGGER.info(f"With scorer=None: {grid_search_cv.best_score_}")

    return grid_search_cv


def custom_cost_to_minimize(
    cost_tpr,
    tpr,
    cost_fpr,
    fpr,
):
    return cost_tpr * (1 - tpr) + cost_fpr * fpr


def find_optimal_threshold_roc(
    y,
    y_pred_proba,
    default_threshold,
    threshold_moving,
    target_tpr,
    target_tnr,
    cost_tpr,
    cost_fpr,
):
    from sklearn.metrics import roc_curve

    # get ROC curve
    fpr, tpr, thresholds_roc = roc_curve(y, y_pred_proba)

    # get thresholds from ROC curve
    df_fpr_tpr = pd.DataFrame({"FPR": fpr, "TPR": tpr, "Threshold": thresholds_roc})
    idx_default = np.argmin(np.abs(thresholds_roc - default_threshold))
    if threshold_moving:
        idx_tpr = np.argmax(tpr >= target_tpr)
        fixed_threshold_tpr = thresholds_roc[idx_tpr]
        tnr = 1 - fpr
        idx_tnr = np.abs(tnr - target_tnr).argmin()
        fixed_threshold_tnr = thresholds_roc[idx_tnr]
        costs = custom_cost_to_minimize(cost_tpr, tpr, cost_fpr, fpr)
        optimal_idx = np.argmin(costs)
        optimal_threshold = thresholds_roc[optimal_idx]

    # generate curve
    plt.figure(figsize=(8, 8))
    fontsize = 12
    plt.scatter(
        df_fpr_tpr["FPR"], df_fpr_tpr["TPR"], s=5, alpha=0.8, color="black", zorder=2
    )

    plt.plot(
        df_fpr_tpr["FPR"],
        df_fpr_tpr["TPR"],
        linestyle="-",
        alpha=0.8,
        color="black",
        zorder=2,
        label=f"{ESTIMATOR_PIPELINE}",
    )

    # No-Skill classifier
    plt.plot(
        [0, 1],
        [0, 1],
        linestyle="--",
        color="gray",
        alpha=0.8,
        label="No-skill classifier",
    )

    # Highlight thresholds point
    plt.scatter(
        fpr[idx_default],
        tpr[idx_default],
        alpha=0.8,
        color="blue",
        s=120,
        marker="o",
        label=f"Default threshold: {default_threshold}",
    )
    if threshold_moving:
        plt.scatter(
            fpr[optimal_idx],
            tpr[optimal_idx],
            alpha=0.8,
            color="green",
            s=120,
            marker="o",
            label=f"Optimal threshold: {optimal_threshold:.3f}",
        )
        plt.scatter(
            fpr[idx_tpr],
            tpr[idx_tpr],
            alpha=0.8,
            color="orange",
            s=120,
            marker="o",
            label=f"TPR≈{target_tpr} threshold: {fixed_threshold_tpr:.3f}",
        )
        plt.scatter(
            fpr[idx_tnr],
            tpr[idx_tnr],
            alpha=0.8,
            color="red",
            s=120,
            marker="o",
            label=f"TNR≈{target_tnr} threshold: {fixed_threshold_tnr:.3f}",
        )

    info_text = (
        f"Info:\n"
        f"        TPR = TP / (TP + FN)\n"
        f"        FPR = FP / (FP + TN)\n"
        f"        TNR = 1 - FPR\n"
        f"        Optimal Cost(TPR, FPR) =\n"
        f"        {cost_tpr} * (1 - TPR) + {cost_fpr} * FPR"
    )
    plt.plot([1, 1], [1, 1], linestyle="", alpha=0.0, label=f"{info_text}")

    plt.legend(fontsize=fontsize, loc="lower right", framealpha=0.6)
    plt.xlim(-0.01, 1)
    plt.ylim(0, 1.01)
    plt.xlabel("False Positive Rate (FPR)", fontsize=fontsize + 2)
    plt.ylabel("True Positive Rate (TPR)", fontsize=fontsize + 2)
    plt.xticks(fontsize=fontsize - 1)
    plt.yticks(fontsize=fontsize - 1)
    plt.title(f"aeid: {AEID}", fontsize=fontsize + 3)

    plt.grid()
    plt.tight_layout()

    plt.savefig(Path(DUMP_FOLDER) / "roc_curve.svg", format="svg")
    plt.close("all")

    # create and store dataframe with thresholds
    if threshold_moving:
        thresholds_list = [
            default_threshold,
            optimal_threshold,
            fixed_threshold_tpr,
            fixed_threshold_tnr,
        ]
        thresholds_df = pd.DataFrame(
            {
                "threshold": thresholds_list,
                "threshold_name": ["default", "optimal", "tpr", "tnr"],
            }
        )
        thresholds_df.round(5).to_csv(Path(DUMP_FOLDER) / "thresholds.csv", index=False)
        LOGGER.info("Optimal and fixed threshold saved.")

        return optimal_threshold, fixed_threshold_tpr, fixed_threshold_tnr

    else:
        return None, None, None


def plot_precision_recall_curve(
    y,
    y_pred_proba,
    default_threshold,
):
    from sklearn.metrics import precision_recall_curve

    # get precision recall curve
    prec, recall, thresholds_pr = precision_recall_curve(y, y_pred_proba)
    #    thresholds_pr = np.append(thresholds_pr, thresholds_pr[-1])

    # create dataframe
    df_pr = pd.DataFrame(
        {"Precision": prec[:-1], "Recall": recall[:-1], "Threshold": thresholds_pr}
    )
    idx_default = np.argmin(np.abs(thresholds_pr - default_threshold))

    # generate curve
    plt.figure(figsize=(8, 8))
    fontsize = 12
    plt.scatter(
        df_pr["Recall"], df_pr["Precision"], s=5, alpha=0.8, color="black", zorder=2
    )

    plt.plot(
        df_pr["Recall"],
        df_pr["Precision"],
        linestyle="-",
        alpha=0.8,
        color="black",
        zorder=2,
        label=f"{ESTIMATOR_PIPELINE}",
    )

    # No-Skill classifier
    npos_rel = len(y[y == 1]) / len(y)
    plt.plot(
        [0, 1],
        [npos_rel, npos_rel],
        linestyle="--",
        color="gray",
        alpha=0.8,
        label="No-skill classifier",
    )

    # Highlight thresholds point
    plt.scatter(
        recall[idx_default],
        prec[idx_default],
        alpha=0.8,
        color="blue",
        s=120,
        marker="o",
        label=f'"Default" threshold: {default_threshold}',
    )

    plt.legend(fontsize=fontsize, loc="lower left", framealpha=0.6)
    plt.xlim(-0.01, 1)
    plt.ylim(0, 1.01)
    plt.xlabel("Recall", fontsize=fontsize + 2)
    plt.ylabel("Precision", fontsize=fontsize + 2)
    plt.xticks(fontsize=fontsize - 1)
    plt.yticks(fontsize=fontsize - 1)
    plt.title(f"aeid: {AEID}", fontsize=fontsize + 3)

    plt.grid()
    plt.tight_layout()

    plt.savefig(Path(DUMP_FOLDER) / "pr_curve.svg", format="svg")
    plt.close("all")


def predict_and_report_classification(
    X,
    y,
    best_estimator,
    default_threshold,
    threshold_moving,
    target_tpr,
    target_tnr,
    cost_tpr,
    cost_fpr,
    cmap="Blues",
):
    from sklearn.metrics import (
        precision_score,
        recall_score,
        f1_score,
        balanced_accuracy_score,
        average_precision_score,
        roc_auc_score,
        matthews_corrcoef,
        classification_report,
        confusion_matrix,
        ConfusionMatrixDisplay,
    )

    # Get the prediction probabilites
    y_pred_proba = best_estimator.predict_proba(X)[:, 1]

    # Get the predictions for the default threshold
    LOGGER.info(f"Default threshold: {default_threshold}")
    y_pred_default_threshold = np.where(y_pred_proba >= default_threshold, 1, 0)

    # Get common metrics
    support = len(y)
    n_pos = np.sum(y)
    f1_micro = f1_score(y, y_pred_default_threshold, average="micro")
    f1_macro = f1_score(y, y_pred_default_threshold, average="macro")
    recall_micro = recall_score(y, y_pred_default_threshold, average="micro")
    recall_macro = recall_score(y, y_pred_default_threshold, average="macro")
    precision_micro = precision_score(y, y_pred_default_threshold, average="micro")
    precision_macro = precision_score(y, y_pred_default_threshold, average="macro")
    balanced_accuracy = balanced_accuracy_score(y, y_pred_default_threshold)
    mcc = matthews_corrcoef(y, y_pred_default_threshold)
    pr_auc = average_precision_score(y, y_pred_proba)
    roc_auc = roc_auc_score(y, y_pred_proba)

    # Save the metrics
    metrics_df = pd.DataFrame(
        {
            "support": [support],
            "n_pos": [n_pos],
            "recall_micro": [recall_micro],
            "recall_macro": [recall_macro],
            "precision_micro": [precision_micro],
            "precision_macro": [precision_macro],
            "f1_micro": [f1_micro],
            "f1_macro": [f1_macro],
            "balanced_accuracy": [balanced_accuracy],
            "mcc": [mcc],
            "pr_auc": [pr_auc],
            "roc_auc": [roc_auc],
        }
    )
    metrics_df.round(5).to_csv(Path(DUMP_FOLDER) / "metrics.csv", index=False)

    # concatenate true values and predictions
    data = {
        "Actual": y,
        "Prediction Probability": y_pred_proba,
        "Prediction Default": y_pred_default_threshold,
    }

    # start lists for predictions
    y_preds = [y_pred_default_threshold]
    y_preds_names = ["default"]
    y_preds_descs = [f"Classification Threshold default={default_threshold}"]

    # Precision recall curve
    plot_precision_recall_curve(
        y,
        y_pred_proba,
        default_threshold=default_threshold,
    )

    # Adjust predictions based on classification threshold
    # ROC curve
    (
        optimal_threshold,
        fixed_threshold_tpr,
        fixed_threshold_tnr,
    ) = find_optimal_threshold_roc(
        y,
        y_pred_proba,
        default_threshold,
        threshold_moving,
        target_tpr=target_tpr,
        target_tnr=target_tnr,
        cost_tpr=cost_tpr,
        cost_fpr=cost_fpr,
    )
    LOGGER.info(f"Optimal threshold: {optimal_threshold}")
    LOGGER.info(f"Fixed threshold TPR≈{target_tpr}: {fixed_threshold_tpr}")
    LOGGER.info(f"Fixed threshold, TNR≈{target_tnr}: {fixed_threshold_tnr}")

    if threshold_moving:
        y_pred_optimal_threshold = np.where(y_pred_proba >= optimal_threshold, 1, 0)
        y_pred_fixed_threshold_tpr = np.where(y_pred_proba >= fixed_threshold_tpr, 1, 0)
        y_pred_fixed_threshold_tnr = np.where(y_pred_proba >= fixed_threshold_tnr, 1, 0)

        y_preds += [
            y_pred_optimal_threshold,
            y_pred_fixed_threshold_tpr,
            y_pred_fixed_threshold_tnr,
        ]
        y_preds_names += ["optimal", "tpr", "tnr"]
        y_preds_descs += [
            "Classification Threshold by cost(TPR, TNR)",
            f"Classification Threshold by TPR≈{target_tpr}",
            f"Classification Threshold by TNR≈{target_tnr}",
        ]

        new_data = {
            "Prediction Optimal": y_pred_optimal_threshold,
            "Prediction TPR": y_pred_fixed_threshold_tpr,
            "Prediction TNR": y_pred_fixed_threshold_tnr,
        }
        data.update(new_data)
        # data = pd.DataFrame(data)

    # df = pd.DataFrame(data)
    # df.round(5).to_csv(os.path.join(DUMP_FOLDER, "estimator_results.csv"), index=False)

    # Confusion Matrix
    labels = [True, False]
    for i, y_pred in enumerate(y_preds):
        LOGGER.info("." * 40)
        LOGGER.info(f"Predict {y_preds_names[i]}")
        name = y_preds_names[i]
        desc = y_preds_descs[i]
        report_dict = classification_report(
            y,
            y_pred,
            labels=labels,
            output_dict=True,
        )
        report_df = pd.DataFrame(report_dict).transpose()
        report_df.round(5).to_csv(Path(DUMP_FOLDER) / f"report_{name}.csv")

        cm = confusion_matrix(
            y,
            y_pred,
            labels=labels,
        )
        tn, fp, fn, tp = cm.ravel()  # Extract values from confusion matrix
        LOGGER.info(f"Total: {len(y)} datapoints")
        LOGGER.info(f"Ground truth: {tp + fn} positive, {tn + fp} negative")
        LOGGER.info(f"Prediction: {tp + fp} positive, {tn + fn} negative")

        display_labels = {
            "Positive": {"fontsize": 30},
            "Negative": {"fontsize": 30},
        }

        plt.figure(figsize=(8, 8))
        cm_display = ConfusionMatrixDisplay(
            confusion_matrix=cm, display_labels=display_labels
        )
        cm_display.plot(cmap=cmap, colorbar=False)

        pos_count = cm[0, 0] + cm[0, 1]
        neg_count = cm[1, 0] + cm[1, 1]

        plt.suptitle(f"Confusion Matrix: {desc} ", fontsize=10)
        plt.title(
            f"aeid: {AEID}, {ESTIMATOR_PIPELINE}, Count: {len(y)} (P:{pos_count}, N:{neg_count})",
            fontsize=10,
        )
        plt.savefig(Path(DUMP_FOLDER) / f"cm_{name}.svg", format="svg")
        plt.close("all")


class ElapsedTimeFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None, style="%"):
        super().__init__(fmt, datefmt, style)
        self.start_time = START_TIME

    def format(self, record):
        delta = datetime.now() - self.start_time
        hours, remainder = divmod(delta.total_seconds(), 3600)
        minutes, seconds = divmod(remainder, 60)
        hundredths = int((delta.microseconds / 10000) % 100)
        elapsed_time_formatted = (
            f"[{int(hours):02}:{int(minutes):02}:{int(seconds):02}:{hundredths:02}]"
        )
        return f"{elapsed_time_formatted} {super().format(record)}"


def get_timestamp(time_point):
    return time_point.strftime("%Y-%m-%d_%H-%M-%S")


def report_exception(exception, traceback_info, entitiy):
    error_file_path = Path(LOG_PATH) / "log/error.error"
    with open(error_file_path, "a") as f:
        err_msg = f"{entitiy} failed: {exception}"
        LOGGER.error(err_msg)
        LOGGER.error(traceback_info)
        print(err_msg, file=f)
        print(traceback_info, file=f)


def save_model(best_estimator, fit_set):
    best_estimator_path = Path(DUMP_FOLDER) / f"best_estimator_{fit_set}.joblib"
    joblib.dump(best_estimator, best_estimator_path, compress=3)


def get_feature_importance_if_applicable(
    best_estimator,
    feature_names,
):
    best_estimator_model = best_estimator[
        -1
    ]  # best_estimator is the pipeline and the last step is the model itself
    try:
        feature_importances = best_estimator_model.feature_importances_
        importance_df = pd.DataFrame(
            {"feature": feature_names, "importances": feature_importances}
        )
        importance_df = importance_df.sort_values(by="importances", ascending=False)
        importance_df.to_csv(
            Path(DUMP_FOLDER) / "sorted_feature_importances.csv", index=False
        )

    except Exception:
        feature_importances = None
        LOGGER.error(f"Feature Importance not implemented for {ESTIMATOR_PIPELINE}")


def _create_empty_log_file(filename):
    with open(filename, "w", encoding="utf-8"):
        pass


def init_logger(logs_folder):
    global LOGGER, LOG_PATH, LOGGER_FOLDER
    # create folder for run
    RUN_FOLDER = get_timestamp(START_TIME)
    LOG_PATH = os.path.join(logs_folder, RUN_FOLDER)
    LOGGER_FOLDER = os.path.join(LOG_PATH, "log")
    os.makedirs(LOGGER_FOLDER, exist_ok=True)
    log_filename = os.path.join(LOGGER_FOLDER, "modelfitting.log")
    error_filename = os.path.join(LOGGER_FOLDER, "modelfitting.error")
    _create_empty_log_file(log_filename)
    _create_empty_log_file(error_filename)

    # setup logger
    file_handler = logging.FileHandler(log_filename, encoding="utf-8")
    console_handler = logging.StreamHandler(stream=sys.stdout)
    formatter = ElapsedTimeFormatter("%(message)s")
    LOGGER.setLevel(logging.INFO)
    file_handler.setFormatter(formatter)
    console_handler.setFormatter(formatter)
    LOGGER.addHandler(file_handler)
    LOGGER.addHandler(console_handler)
    return LOGGER, RUN_FOLDER


def add_status_file(status):
    with open(os.path.join(DUMP_FOLDER, f"{status}.txt"), "w") as file:
        file.write(status)


def get_total_elapsed_time():
    elapsed_seconds = round((datetime.now() - START_TIME).total_seconds(), 2)
    hours, remainder = divmod(elapsed_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    elapsed_formatted = f"{int(hours):02}:{int(minutes):02}:{int(seconds):02}"
    return elapsed_formatted
